import matplotlib.pyplot as plt
import numpy as np
from IPython.core.getipython import get_ipython
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from autoemulate.core.types import NumpyLike


def display_figure(fig: Figure):
    """
    Display a matplotlib figure.

    Display a matplotlib figure appropriately based on the environment
    (Jupyter notebook or terminal).

    Parameters
    ----------
    fig: Figure
        The object to display.

    Returns
    -------
    Figure
        The input figure object.
    """
    # Are we in Jupyter?
    try:
        is_jupyter = get_ipython().__class__.__name__ == "ZMQInteractiveShell"
    except NameError:
        is_jupyter = False

    if is_jupyter:
        # we don't show otherwise it will double plot
        plt.close(fig)
        return fig
    # in terminal, show the plot
    plt.close(fig)
    plt.show()
    return fig


def plot_xy(  # noqa: PLR0913
    x: NumpyLike,
    y: NumpyLike,
    y_pred: NumpyLike,
    y_variance: NumpyLike | None = None,
    ax: Axes | None = None,
    title: str = "xy",
    input_index: int | None = None,
    output_index: int | None = None,
    r2_score: float | None = None,
):
    """
    Plot observed and predicted values vs. features.

    Parameters
    ----------
    x: NumpyLike
        An array of inputs.
    y: NumpyLike
        An array of outputs.
    y_pred: NumpyLike
        An array of predictions.
    y_variance: NumpyLike | None
        An optional array of predictive variances.
    ax: Axes | None
        An optional matplotlib Axes object to plot on.
    title: str
        An optional title for the plot.
    input_index: int | None
        An optional index of the input dimension to plot.
    output_index: int | None
        An optional index of the output dimension to plot.
    r2_score: float | None
        An option r2 score to include in the plot legend.
    """
    # Sort the data
    sort_idx = np.argsort(x).flatten()
    x_sorted = x[sort_idx]
    y_sorted = y[sort_idx]
    y_pred_sorted = y_pred[sort_idx]
    y_std = None
    if y_variance is not None:
        y_variance_sorted = y_variance[sort_idx]
        y_std = np.sqrt(y_variance_sorted)

    org_points_color = "Goldenrod"
    pred_points_color = "#6A5ACD"
    pred_line_color = "#6A5ACD"
    ci_color = "lightblue"

    assert ax is not None, "ax must be provided"
    if y_std is not None:
        ax.fill_between(
            x_sorted,
            y_pred_sorted - 2 * y_std,
            y_pred_sorted + 2 * y_std,
            color=ci_color,
            alpha=0.25,
            label="95% Confidence Interval",
        )
    ax.plot(
        x_sorted,
        y_pred_sorted,
        color=pred_line_color,
        label="pred.",
        alpha=0.8,
        linewidth=1,
    )  # , linestyle='--'
    ax.scatter(
        x_sorted,
        y_sorted,
        color=org_points_color,
        alpha=0.7,
        edgecolor="black",
        linewidth=0.5,
        label="data",
    )
    ax.scatter(
        x_sorted,
        y_pred_sorted,
        color=pred_points_color,
        edgecolor="black",
        linewidth=0.5,
        alpha=0.7,
        label="pred.",
    )

    ax.set_xlabel(f"$x_{input_index}$", fontsize=13)
    ax.set_ylabel(f"$y_{output_index}$", fontsize=13)
    ax.set_title(title, fontsize=13)
    ax.grid(True, alpha=0.3)

    # Get the handles and labels for the scatter plots
    handles, _ = ax.get_legend_handles_labels()

    # Add legend and get its bounding box
    lbl = "pred." if y_variance is None else "pred. (±2σ)"  # noqa: RUF001
    legend = ax.legend(
        handles[-2:],
        ["data", lbl],
        loc="best",
        handletextpad=0,
        columnspacing=0,
        ncol=2,
    )

    # Place R² just below the legend
    if legend:
        # Get the bounding box of the legend in axes coordinates
        bbox = legend.get_window_extent(ax.figure.canvas.get_renderer())  # pyright: ignore[reportAttributeAccessIssue]
        inv = ax.transAxes.inverted()
        bbox_axes = bbox.transformed(inv)
        # Place the text just below the legend
        text_x = bbox_axes.x0
        text_y = bbox_axes.y0 - 0.04  # small offset below legend
        ax.text(
            text_x,
            text_y,
            f"R\u00b2 = {r2_score:.6f}",
            transform=ax.transAxes,
            verticalalignment="top",
        )
    else:
        # fallback: place in lower left
        ax.text(
            0.05,
            0.05,
            f"R\u00b2 = {r2_score:.6f}",
            transform=ax.transAxes,
            verticalalignment="bottom",
        )


def calculate_subplot_layout(n_plots, n_cols=3):
    """
    Calculate optimal number of rows and columns for subplots.

    Parameters
    ----------
    n_plots: int
        Number of plots to display.
    n_cols: int
        Maximum number of columns allowed. Defaults to 3.

    Returns
    -------
    tuple
        (n_rows, n_cols) for the subplot layout.
    """
    if n_plots <= 1:
        return (1, 1)

    n_cols = min(n_plots, n_cols)
    n_rows = (n_plots + n_cols - 1) // n_cols

    return n_rows, n_cols



import logging

import torch
import torch.nn.functional as F
from torch import nn
from torch.distributions import Transform, constraints

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.data.utils import set_random_seed
from autoemulate.transforms.base import AutoEmulateTransform


class VAE(nn.Module, TorchDeviceMixin):
    """
    Variational Autoencoder implementation in PyTorch.

    Parameters
    ----------
    input_dim: int
        Dimensionality of input data.
    hidden_layers: list
        List of hidden dimensions for encoder and decoder networks.
    latent_dim: int
        Dimensionality of the latent space.
    device: DeviceLike | None
        Device to run the model on. If None, uses the default device (usually CPU or
        GPU). Defaults to None.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_layers: list,
        latent_dim: int,
        device: DeviceLike | None = None,
    ):
        nn.Module.__init__(self)
        TorchDeviceMixin.__init__(self, device=device)

        # Encoder layers
        encoder_layers = []
        prev_dim = input_dim
        for dim in hidden_layers:
            encoder_layers.append(nn.Linear(prev_dim, dim, device=self.device))
            encoder_layers.append(nn.ReLU())
            prev_dim = dim

        self.encoder = nn.Sequential(*encoder_layers)
        self.fc_mu = nn.Linear(hidden_layers[-1], latent_dim, device=self.device)
        self.fc_var = nn.Linear(hidden_layers[-1], latent_dim, device=self.device)

        # Decoder layers
        decoder_layers = []
        prev_dim = latent_dim
        for dim in reversed(hidden_layers):
            decoder_layers.append(nn.Linear(prev_dim, dim, device=self.device))
            decoder_layers.append(nn.ReLU())
            prev_dim = dim
        decoder_layers.append(nn.Linear(prev_dim, input_dim, device=self.device))

        self.decoder = nn.Sequential(*decoder_layers)

    def encode(self, x):
        """Encode input to mean and log variance of latent distribution."""
        x = self.encoder(x)
        mu = self.fc_mu(x)
        log_var = self.fc_var(x)
        return mu, log_var

    def reparameterize(self, mu, log_var):
        """Sample from latent distribution using reparameterization trick."""
        std = torch.exp(0.5 * log_var)
        eps = torch.randn_like(std)
        return mu + eps * std

    def decode(self, z):
        """Decode latent representation back to original space."""
        return self.decoder(z)

    def forward(self, x):
        """Full forward pass through the VAE."""
        mu, log_var = self.encode(x)
        z = self.reparameterize(mu, log_var)
        recon = self.decode(z)
        return recon, mu, log_var

    def loss_function(self, recon_x, x, mu, log_var, beta=1.0):
        """Calculate VAE loss: reconstruction + KL divergence."""
        # MSE reconstruction loss
        recon_loss = F.mse_loss(recon_x, x, reduction="sum")

        # KL divergence
        kl_loss = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())

        return recon_loss + beta * kl_loss


class VAETransform(AutoEmulateTransform):
    """VAE transform for dimensionality reduction."""

    domain = constraints.real
    codomain = constraints.real
    bijective = False
    vae: VAE | None = None

    def __init__(  # noqa: PLR0913
        self,
        latent_dim: int = 3,
        hidden_layers: list[int] | None = None,
        epochs: int = 800,
        batch_size: int = 32,
        learning_rate: float = 1e-3,
        random_seed: int | None = None,
        beta: float = 1.0,
        verbose: bool = False,
        cache_size: int = 0,
    ):
        """
        Initialize the VAE transform parameters.

        Intialize the VAE transform parameters but defer intialization of the inner
        VAE model until fit is called when the input data is available.

        Parameters
        ----------
        latent_dim: int
            The dimensionality of the VAE latent space. Defaults to 3.
        hidden_layers: list of int
            The number of hidden layers and their sizes in the VAE. If None, defaults to
            [64, 32]. Defaults to None.
        epochs: int
            The number of training epochs for the VAE. Defaults to 800.
        batch_size: int
            The batch size for training the VAE. Defaults to 32.
        learning_rate: float
            The learning rate for the VAE optimizer. Defaults to 1e-3.
        random_seed: int
            Random seed for reproducibility. Defaults to None.
        beta: float
            The beta parameter for the VAE loss function, controlling the trade-off
            between reconstruction loss and KL divergence. Defaults to 1.0.
        verbose: bool
            If True, log training progress. Defaults to False.
        cache_size: int
            Whether to cache previous transform. Set to 0 to disable caching. Set to
            1 to enable caching of the last single value. This might be useful for
            repeated expensive calls with the same input data but is by default
            disabled. See `PyTorch documentation <https://github.com/pytorch/pytorch/blob/134179474539648ba7dee1317959529fbd0e7f89/torch/distributions/transforms.py#L46-L89>`_
            for more details on caching. Defaults to 0.
        """
        Transform.__init__(self, cache_size=cache_size)
        self.latent_dim = latent_dim
        self.hidden_layers = hidden_layers or [64, 32]
        self.epochs = epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.beta = beta
        self.random_seed = random_seed
        self.verbose = verbose
        self.cache_size = cache_size  # Store for serialization

        # Initialized during fit
        self.vae = None
        self.input_dim = None
        self._is_fitted = False

    def _init_vae(self, intput_dim: int):
        self.input_dim = intput_dim
        self.vae = VAE(
            intput_dim, self.hidden_layers, self.latent_dim, device=self.device
        ).to(self.device)

    def fit(self, x: TensorLike):
        """Fit the VAE on the training data."""
        TorchDeviceMixin.__init__(self, device=x.device)

        # Set random seed for reproducibility
        if self.random_seed is not None:
            set_random_seed(self.random_seed)

        # Create dataloader
        data_loader = self._convert_to_dataloader(
            x, x, batch_size=self.batch_size, shuffle=True
        )

        # Initialize the model
        self._init_vae(intput_dim=x.shape[1])
        assert self.vae is not None

        # Train the VAE
        optimizer = torch.optim.Adam(self.vae.parameters(), lr=self.learning_rate)
        self.vae.train()
        for epoch in range(self.epochs):
            total_loss = 0
            for batch_x, _ in data_loader:
                recon_batch, mu, log_var = self.vae(batch_x)
                loss = self.vae.loss_function(
                    recon_batch, batch_x, mu, log_var, beta=self.beta
                )
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                total_loss += loss.item()

            # Log progress
            if self.verbose and (epoch + 1) % 10 == 0:
                msg = (
                    f"Epoch {epoch + 1}/{self.epochs}, "
                    f"Loss: {total_loss / len(data_loader):.4f}"
                )
                logging.info(msg)

        self._is_fitted = True

    def _call(self, x):
        self._check_is_fitted()
        assert self.vae is not None
        self.vae.eval()
        with torch.no_grad():
            mu, _ = self.vae.encode(x)
            return mu

    def _inverse(self, y):
        self._check_is_fitted()
        assert self.vae is not None
        self.vae.eval()
        with torch.no_grad():
            return self.vae.decode(y)

    def log_abs_det_jacobian(self, x, y):
        """Log abs det Jacobian not computable since transform is not bijective."""
        _, _ = x, y
        msg = "log det Jacobian not computable since transform is not bijective."
        raise RuntimeError(msg)

    def _jacobian(self, y: TensorLike) -> TensorLike:
        # Delta method to computing covariance in original space of transform's domain
        # requires the calculation of the jacobian of the transform.
        # https://github.com/alan-turing-institute/autoemulate/issues/376#issuecomment-2891374970
        self._check_is_fitted()
        assert self.vae is not None
        # Ensure the input tensor requires gradient for jacobian computation
        if not y.requires_grad:
            y = y.detach().clone().requires_grad_(True)
        jacobian = torch.autograd.functional.jacobian(self.vae.decode, y)
        assert isinstance(jacobian, TensorLike)
        return jacobian

    def _batch_basis_matrix(self, y):
        n = y.shape[0]
        jacobian = self._jacobian(y)

        # Reshape jacobian to a batch local basis matrices by stacking along the diag
        return torch.stack([jacobian[i, :, i, :] for i in range(n)], 0)

    def _expanded_basis_matrix(self, y):
        jacobian = self._jacobian(y)
        n = y.shape[0]

        # Reshape jacobian for shape of cov_y (n_tasks x n_samples)
        return jacobian.reshape(n * jacobian.shape[1], -1)



import torch
from torch.distributions import Transform, constraints

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import TensorLike
from autoemulate.transforms.base import AutoEmulateTransform


class PCATransform(AutoEmulateTransform):
    """PCA transform for dimensionality reduction."""

    domain = constraints.real
    codomain = constraints.real
    bijective = False

    def __init__(self, n_components: int, niter: int = 1000, cache_size: int = 0):
        """
        Initialize the PCA transform.

        Parameters
        ----------
        n_components: int
            The number of principal components to use for the transformation.
        niter: int
            The number of iterations for the PCA algorithm. Defaults to 1000.
        cache_size: int
            Whether to cache previous transform. Set to 0 to disable caching. Set to
            1 to enable caching of the last single value. This might be useful for
            repeated expensive calls with the same input data but is by default
            disabled. See `PyTorch documentation <https://github.com/pytorch/pytorch/blob/134179474539648ba7dee1317959529fbd0e7f89/torch/distributions/transforms.py#L46-L89>`_
            for more details on caching. Defaults to 0.
        """
        Transform.__init__(self, cache_size=cache_size)
        self.n_components = n_components  # n_c
        self.niter = niter
        self.cache_size = cache_size  # Store for serialization

    def fit(self, x: TensorLike):
        """Fit the PCA transform to the data."""
        TorchDeviceMixin.__init__(self, device=x.device)
        self.check_matrix(x)
        self.mean = x.mean(0, keepdim=True)  # (1, d)
        _, _, v = torch.pca_lowrank(x, q=self.n_components, niter=self.niter)
        self.components = v[:, : self.n_components]  # (d, n_c)
        self._is_fitted = True

    def _call(self, x: TensorLike):
        self._check_is_fitted()
        return (x - self.mean) @ self.components

    def _inverse(self, y: TensorLike):
        self._check_is_fitted()
        return y @ self.components.T + self.mean  # (n, n_c) x (n_c, d) + (1, d)

    def log_abs_det_jacobian(self, x: TensorLike, y: TensorLike):
        """Log abs det Jacobian not computable for n_components < d as not bijective."""
        _, _ = x, y
        msg = (
            "log abs det Jacobian not computable for n_components < d as not bijective."
        )
        raise RuntimeError(msg)

    @property
    def _basis_matrix(self) -> TensorLike:
        self._check_is_fitted()
        return self.components



from typing import Literal

import numpy as np
from sklearn.svm import SVR
from sklearn.utils.validation import check_X_y

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, NumpyLike, TensorLike
from autoemulate.emulators.base import SklearnBackend
from autoemulate.transforms.standardize import StandardizeTransform


class SupportVectorMachine(SklearnBackend):
    """
    Support Vector Machines Emulator.

    Wraps Support Vector Regressor from scikit-learn.
    """

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike,
        y: TensorLike | None,
        standardize_x: bool = False,
        standardize_y: bool = False,
        kernel: Literal["linear", "poly", "rbf", "sigmoid", "precomputed"] = "rbf",
        degree: int = 3,
        gamma: Literal["scale", "auto"] | float = "scale",
        coef0: float = 0.0,
        tol: float = 1e-3,
        C: float = 1.0,
        epsilon: float = 0.1,
        shrinking: bool = True,
        cache_size: float = 200.0,
        verbose: bool = False,
        max_iter: int = 100,
        normalise_y: bool = True,
        device: DeviceLike = "cpu",
    ):
        """Initialize a SupportVectorMachine emulator.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike | None
            Target values. If None, the model will be fitted later.
        standardize_x: bool
            Whether to standardize input features. Defaults to False.
        standardize_y: bool
            Whether to standardize target values. Defaults to False.
        kernel: {"linear", "poly", "rbf", "sigmoid", "precomputed"}
            Kernel type to be used in the algorithm. Options are:
            "linear", "poly", "rbf", "sigmoid", "precomputed". Defaults to "rbf".
        degree: int
            Degree of the polynomial kernel function ('poly'). Ignored by other kernels.
            Defaults to 3.
        gamma: {"scale", "auto"} | float
            Kernel coefficient for 'rbf', 'poly', and 'sigmoid'. If 'scale', it is set
            to 1 / (n_features * X.var()) as default; if 'auto', it is set to
            1 / n_features. If float, must be non-negative.
            Defaults to 'scale'.
        coef0: float
            Independent term in kernel function. It is only significant in 'poly' and
            'sigmoid'. Defaults to 0.0.
        tol: float
            Tolerance for stopping criteria. Defaults to 1e-3.
        C: float
            Regularization parameter. The strength of the regularization is inversely
            proportional to C. Must be strictly positive. Defaults to 1.0.
        epsilon: float
            Epsilon in the epsilon-SVR model. It specifies the epsilon-tube within which
            no penalty is associated in the training loss function with points predicted
            within a distance epsilon from the actual value. Defaults to 0.1.
        shrinking: bool
            Whether to use the shrinking heuristic. Defaults to True.
        cache_size: float
            Specify the size of the kernel cache (in MB). Defaults to 200.
        verbose: bool
            Enable verbose output. Defaults to False.
        max_iter: int
            Hard limit on iterations within solver, or -1 for no limit. Defaults to 100.
        normalise_y: bool
            Whether to normalize the target values y before fitting the model. Defaults
            to True.
        device: DeviceLike
            Device to run the model on (e.g., "cpu", "cuda", "mps"). Note: SVR
            always runs on CPU, this parameter is for compatibility with the base class.
        """
        _, _, _ = x, y, device  # ignore unused arguments
        TorchDeviceMixin.__init__(self, device=device, cpu_only=True)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.kernel = kernel
        self.degree = degree
        self.gamma = gamma
        self.coef0 = coef0
        self.tol = tol
        self.C = C
        self.epsilon = epsilon
        self.shrinking = shrinking
        self.cache_size = cache_size
        self.verbose = verbose
        self.max_iter = max_iter
        self.normalise_y = normalise_y
        self.model = SVR(
            kernel=self.kernel,
            degree=self.degree,
            gamma=self.gamma,  # type: ignore reportArgumentType
            coef0=self.coef0,
            tol=self.tol,
            C=self.C,
            epsilon=self.epsilon,
            shrinking=self.shrinking,
            cache_size=self.cache_size,  # type: ignore reportArgumentType
            verbose=self.verbose,
            max_iter=self.max_iter,
        )

    @staticmethod
    def is_multioutput() -> bool:
        """Support Vector Machines do not support multi-output."""
        return False

    def _model_specific_check(self, x: NumpyLike, y: NumpyLike):
        check_X_y(x, y, ensure_min_samples=2)

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        return {
            "kernel": ["rbf", "linear", "poly", "sigmoid"],
            "degree": [np.random.randint(2, 6)],
            "gamma": ["scale", "auto"],
            "coef0": [np.random.uniform(0.0, 1.0)],
            "tol": [np.random.uniform(1e-5, 1e-3)],
            "C": [np.random.uniform(1.0, 3.0)],
            "epsilon": [np.random.uniform(0.1, 0.3)],
            "shrinking": [True, False],
            "max_iter": [-1],
        }



from abc import abstractmethod
from dataclasses import dataclass, field

import numpy as np
import torch
from tqdm import tqdm

from autoemulate.data.utils import set_random_seed

from ..core.types import GaussianLike, TensorLike
from .base import Active


@dataclass(kw_only=True)
class Stream(Active):
    """
    Active learning strategy for streaming data.

    Parameters
    ----------
    (Inherits parameters from Active)
    """

    show_progress: bool = field(default=True)

    @abstractmethod
    def query(
        self, x: TensorLike | None = None
    ) -> tuple[TensorLike | None, TensorLike | GaussianLike, dict[str, float]]:
        """
        Abstract method to query new samples from a stream.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.

        Returns
        -------
        Tuple[torch.Tensor or None, torch.Tensor, torch.Tensor, Dict[str, List[Any]]]
            A tuple containing:
            - The queried samples (or None if no query is made),
            - The predicted outputs,
            - The covariance estimates,
            - A dictionary of additional metrics.
        """

    def fit_samples(self, x: torch.Tensor):
        """
        Fit the active learner using a stream of samples.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.
        """
        x = self.check_matrix(x)
        for xi in (
            pb := tqdm(
                x,
                desc=self.__class__.__name__,
                leave=True,
                disable=not self.show_progress,
            )
        ):
            self.fit(xi.reshape(1, -1))
            pb.set_postfix(
                ordered_dict={key: val[-1] for key, val in self.metrics.items()}
            )

    def fit_batches(self, x: torch.Tensor, batch_size: int):
        """
        Fit the active learner using batches of samples.

        This method automatically splits x into batches of the specified size and then
        sequentially fits each batch.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.
        batch_size: int
            Number of samples per batch.
        """
        x = self.check_matrix(x)
        for i in (
            pb := tqdm(
                range(0, x.shape[0], batch_size),
                desc=f"{self.__class__.__name__} (batches)",
                disable=not self.show_progress,
            )
        ):
            batch = x[i : i + batch_size]
            self.fit(batch)
            pb.set_postfix(ordered_dict={k: v[-1] for k, v in self.metrics.items()})


@dataclass(kw_only=True)
class Random(Stream):
    """
    Random active learning strategy that queries samples based on a fixed probability.

    Parameters
    ----------
    p_query: float
        Query probability for selecting a sample.
    random_seed: int | None
        Optional random seed for reproducibility.
    """

    p_query: float

    def query(
        self,
        x: TensorLike | None = None,
        random_seed: int | None = None,
    ) -> tuple[torch.Tensor | None, TensorLike | GaussianLike, dict[str, float]]:
        """
        Query new samples randomly based on a fixed probability.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.

        Returns
        -------
        Tuple[torch.Tensor or None, torch.Tensor, torch.Tensor, Dict[str, List[Any]]]
            A tuple containing:
            - The queried samples (or None if the random condition is not met),
            - The predicted outputs,
            - The covariance estimates,
            - An empty dictionary of additional metrics.
        """
        assert isinstance(x, TensorLike)
        # TODO: move handling to check method in base class
        output = self.emulator.predict(x)
        assert isinstance(output, TensorLike | GaussianLike)
        # assert isinstance(output, TensorLike | DistributionLike)
        if random_seed is not None:
            set_random_seed(seed=random_seed)
        x = x if np.random.rand() < self.p_query else None
        return x, output, {}


@dataclass(kw_only=True)
class Threshold(Stream):
    """
    Threshold-based active learning strategy.

    Threshold-based active learning strategy that queries samples based on a score
    threshold.

    Parameters
    ----------
    threshold : float
        Threshold value for querying.
    """

    threshold: float

    def __post_init__(self):
        """
        Initialize the threshold-based learner and update metrics.

        Parameters
        ----------
        x_train: torch.Tensor
            Training input tensor.
        y_train: torch.Tensor
            Training output tensor.
        """
        super().__post_init__()
        self.metrics.update({"score": []})

    @abstractmethod
    def score(
        self, x: torch.Tensor, y: torch.Tensor, Sigma: torch.Tensor
    ) -> torch.Tensor:
        """
        Abstract method to compute a score for querying based on the input sample.

        Parameters
        ----------
        x: torch.Tensor
            Input sample tensor.
        y: torch.Tensor
            Predicted output tensor.
        Sigma: torch.Tensor
            Covariance estimates tensor.

        Returns
        -------
        torch.Tensor
            Computed score for the input sample.
        """

    def query(
        self, x: TensorLike | None = None
    ) -> tuple[
        torch.Tensor | None,
        torch.Tensor | GaussianLike,
        dict[str, float],
    ]:
        """
        Query new samples based on whether the computed score exceeds a threshold.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.

        Returns
        -------
        Tuple[torch.Tensor or None, torch.Tensor, torch.Tensor, Dict[str, List[Any]]]
            A tuple containing:
            - The queried samples (or None if the score does not exceed the threshold),
            - The predicted outputs,
            - The covariance estimates,
            - A dictionary with the computed score.
        """
        # TODO: move handling to check method in base class
        assert isinstance(x, torch.Tensor)
        output = self.emulator.predict(x)
        assert isinstance(output, GaussianLike)
        assert isinstance(output.variance, torch.Tensor)
        score = self.score(x, output.mean, output.variance)
        x = x if score > self.threshold else None
        return x, output, {"score": score.item()}


@dataclass(kw_only=True)
class Input(Threshold):
    """
    Active learning strategy based on input space criteria.

    Parameters
    ----------
    (Inherits parameters from Threshold)
    """


@dataclass(kw_only=True)
class Distance(Input):
    """
    Active learning strategy that scores samples based on distance metrics.

    Queries samples based on the average minimum distance to the training set.

    Parameters
    ----------
    (Inherits parameters from Input)
    """

    def score(
        self, x: torch.Tensor, y: torch.Tensor, Sigma: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute average minimum distance from the input samples to the training data.

        Parameters
        ----------
        x: torch.Tensor
            Input samples.

        Returns
        -------
        float
            The average minimum distance.
        """
        _, _, _ = x, y, Sigma  # Unused variables
        distances = torch.cdist(x, self.x_train)
        min_dists, _ = distances.min(dim=1)
        return min_dists.mean()


@dataclass(kw_only=True)
class Output(Threshold):
    """
    Active learning strategy based on output space criteria.

    Parameters
    ----------
    (Inherits parameters from Threshold)
    """


@dataclass(kw_only=True)
class A_Optimal(Output):
    """
    Active learning strategy using the A-optimal design criterion.

    Uses the trace of the covariance matrix as the scoring function.

    Parameters
    ----------
    (Inherits parameters from Output)
    """

    def score(
        self, x: torch.Tensor, y: torch.Tensor, Sigma: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the score using the trace of the covariance matrix.

        Parameters
        ----------
        x: torch.Tensor
            Input samples.
        y: torch.Tensor
            Predicted outputs (not used).
        Sigma: torch.Tensor
            Covariance estimates (not used).

        Returns
        -------
        torch.Tensor
            Score based on the trace of the covariance matrix.
        """
        _, _ = x, y  # Unused variables
        return self.trace(Sigma, self.out_dim)


@dataclass(kw_only=True)
class D_Optimal(Output):
    """
    Active learning strategy using the D-optimal design criterion.

    Uses the log-determinant of the covariance matrix as the scoring function.

    Parameters
    ----------
    (Inherits parameters from Output)
    """

    def score(
        self, x: torch.Tensor, y: torch.Tensor, Sigma: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the score using the log-determinant of the covariance matrix.

        Parameters
        ----------
        x: torch.Tensor
            Input samples.
        y: torch.Tensor
            Predicted outputs.
        Sigma: torch.Tensor
            Covariance estimates.

        Returns
        -------
        torch.Tensor
            Score based on the log-determinant of the covariance matrix.
        """
        _, _ = x, y  # Unused variables
        return self.logdet(Sigma, self.out_dim)


@dataclass(kw_only=True)
class E_Optimal(Output):
    """
    Active learning strategy using the E-optimal design criterion.

    Uses the maximum eigenvalue of the covariance matrix as the scoring function.

    Parameters
    ----------
    (Inherits parameters from Output)
    """

    def score(
        self, x: torch.Tensor, y: torch.Tensor, Sigma: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the score using the maximum eigenvalue of the covariance matrix.

        Parameters
        ----------
        x: torch.Tensor
            Input samples.
        y: torch.Tensor
            Predicted outputs.
        Sigma: torch.Tensor
            Covariance estimates.

        Returns
        -------
        torch.Tensor
            Score based on the maximum eigenvalue of the covariance matrix.
        """
        _, _ = x, y  # Unused variables
        return self.max_eigval(Sigma)


@dataclass(kw_only=True)
class Adaptive(Threshold):
    """
    Adaptive active learning strategy.

    Learning strategy that adjusts the query threshold using PID control.

    The adaptive mechanism modifies the threshold based on proportional, integral, and
    derivative components. Errors can be computed using a sliding window to reduce
    dependency on initial errors.

    Parameters
    ----------
    Kp: float
        Proportional gain.
    Ki: float
        Integral gain.
    Kd: float
        Derivative gain.
    key: str
        Key used to reference a metric.
    target: float
        Target value for the metric.
    min_threshold: float or None
        Minimum allowable threshold.
    max_threshold: float or None
        Maximum allowable threshold.
    window_size: int or None, optional
        Size of the sliding window to use for error computation. If None, use the full
        error history.
    """

    Kp: float
    Ki: float
    Kd: float
    key: str
    target: float
    min_threshold: float | None
    max_threshold: float | None
    window_size: int | None

    def __post_init__(self):
        """
        Initialize the adaptive learner and update metrics.

        Parameters
        ----------
        x_train: torch.Tensor
            Training input tensor.
        y_train: torch.Tensor
            Training output tensor.
        """
        super().__post_init__()
        self.metrics.update(
            {
                "threshold": [],
                "error_prop": [],
                "error_int": [],
                "error_deriv": [],
            }
        )

    def query(
        self, x: TensorLike | None = None
    ) -> tuple[TensorLike | None, TensorLike | GaussianLike, dict[str, float]]:
        """
        Query new samples based on the adaptive threshold.

        Query new samples and adapt the threshold using a PID controller with a sliding
        window for error computation.

        Parameters
        ----------
        x: torch.Tensor
            Stream of input samples.

        Returns
        -------
        Tuple[torch.Tensor or None, torch.Tensor, torch.Tensor, Dict[str, List[Any]]]
            A tuple containing:
            - The queried samples,
            - The predicted outputs,
            - The covariance estimates,
            - A dictionary with the PID control metrics.
        """
        # Call the parent query (e.g., from Threshold) to get initial values and
        # metrics.
        x, output, metrics = super().query(x)

        # Retrieve the error history for the specified metric.
        error_list = self.metrics.get(self.key, [])
        if error_list:
            if self.window_size is not None:
                n = min(len(error_list), self.window_size)
                errors = torch.tensor(error_list[-n:]) - self.target
            else:
                errors = torch.tensor(error_list) - self.target
        else:
            errors = torch.tensor([0.0])

        # Compute PID components.
        ep = errors[-1].item() if len(errors) >= 1 else 0.0
        ei = errors.sum().item() if len(errors) >= 1 else 0.0
        ed = errors[-1].item() - errors[-2].item() if len(errors) >= 2 else 0.0

        # Update the threshold using the PID control law and enforce bounds.
        self.threshold += self.Kp * ep + self.Ki * ei + self.Kd * ed
        if self.min_threshold is not None:
            self.threshold = max(self.threshold, self.min_threshold)
        if self.max_threshold is not None:
            self.threshold = min(self.threshold, self.max_threshold)

        # Update the metrics dictionary with the PID control terms.
        metrics.update(
            {
                "threshold": self.threshold,
                "error_prop": ep,
                "error_int": ei,
                "error_deriv": ed,
            }
        )
        return x, output, metrics


@dataclass(kw_only=True)
class Adaptive_Distance(Adaptive, Distance):
    """
    Adaptive input distance-based active learning strategy using PID control.

    Parameters
    ----------
    (Inherits parameters from Adaptive and Distance)
    """


@dataclass(kw_only=True)
class Adaptive_A_Optimal(Adaptive, A_Optimal):
    """
    Adaptive A-optimal active learning strategy using PID control.

    Parameters
    ----------
    (Inherits parameters from Adaptive and A_Optimal)
    """

    def __post_init__(self):  # noqa: D105
        if self.min_threshold is not None and self.min_threshold < 0.0:
            msg = (
                f"Minimum threashold ({self.min_threshold}) must be greater than or "
                "equal to 0 since it uses trace of a positive semi-definite matrix."
            )
            raise ValueError(msg)

        if self.min_threshold is None:
            self.min_threshold = 0.0

        A_Optimal.__post_init__(self)


@dataclass(kw_only=True)
class Adaptive_D_Optimal(Adaptive, D_Optimal):
    """
    Adaptive D-optimal active learning strategy using PID control.

    Parameters
    ----------
    (Inherits parameters from Adaptive and D_Optimal)
    """


@dataclass(kw_only=True)
class Adaptive_E_Optimal(Adaptive, E_Optimal):
    """
    Adaptive E-optimal active learning strategy using PID control.

    Parameters
    ----------
    (Inherits parameters from Adaptive and E_Optimal)
    """

    def __post_init__(self):  # noqa: D105
        if self.min_threshold is not None and self.min_threshold < 0.0:
            msg = (
                f"Minimum threshold ({self.min_threshold}) must be greater than or "
                "equal to 0 since it uses max eigenvalue of a positive semi-definite "
                "matrix."
            )
            raise ValueError(msg)

        if self.min_threshold is None:
            self.min_threshold = 0.0

        E_Optimal.__post_init__(self)



import itertools

import torch

from autoemulate.core.types import TensorLike


class PolynomialFeatures:
    """
    Class to map existing feature set to a polynomial feature set.

    Parameters
    ----------
    degree: int
        The degree of the polynomial for which we are defining
        the mapping.
    input_size: int
        The number of features to be mapped.

    Examples
    --------
    Initialize the class to map the feature `X` (`n1` samples x `n2` features):

    >>> pf = PolynomialFeatures(degree=2, input_size=X.shape[1])

    Fit the instance in order to predefine the features that need to be multiplied to
    create the new features.

    >>> pf.fit()

    Generate the new polynomial feature set:

    >>> X_deg_2 = pf.transform(X)
    """

    def __init__(self, degree: int, input_size: int):
        assert degree > 0, "`degree` input must be greater than 0."
        assert input_size > 0, (
            "`input_size`, which defines the number of features, for has to be "
            "greater than 0"
        )
        self.degree = degree
        self.indices = None
        self.input_size = input_size

    def fit(self):
        """Generate the feature combinations for the transformation."""
        x = list(range(self.input_size))

        d = self.degree
        L = []
        while d > 1:
            combinations = [list(p) for p in itertools.product(x, repeat=d)]
            for li in combinations:
                li.sort()
            L += list(map(list, torch.unique(torch.tensor(combinations), dim=0)))
            d -= 1
        L += [[i] for i in x]

        Ls = []
        for d in range(1, self.degree + 1):
            ld = []
            for combination in L:
                if len(combination) == d:
                    ld.append(combination)
            ld.sort()
            Ls += ld
        self.indices = Ls

    def transform(self, x: TensorLike):
        """Generate the polynomial feature set."""
        if not self.indices:
            msg = "self.indices is set to None. Did you forget to call 'fit'?"
            raise ValueError(msg)

        return torch.stack(
            [torch.prod(x[..., i], dim=-1) for i in self.indices], dim=-1
        )



from typing import Literal

import numpy as np
from sklearn.ensemble import RandomForestRegressor

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.emulators.base import SklearnBackend
from autoemulate.transforms.standardize import StandardizeTransform


class RandomForest(SklearnBackend):
    """
    Random forest Emulator.

    Implements Random Forests regression from scikit-learn.
    """

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = False,
        standardize_y: bool = False,
        n_estimators: int = 100,
        criterion: Literal[
            "squared_error", "absolute_error", "friedman_mse", "poisson"
        ] = "squared_error",
        max_depth: int | None = None,
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        max_features: Literal["sqrt", "log2"] | int | float = 1.0,
        bootstrap: bool = True,
        oob_score: bool = False,
        max_samples: int | None = None,
        random_seed: int | None = None,
        device: DeviceLike = "cpu",
        **kwargs,
    ):
        """Initialize a RandomForest emulator.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike
            Target values.
        standardize_x: bool
            Whether to standardize input features. Defaults to False.
        standardize_y: bool
            Whether to standardize target values. Defaults to False.
        n_estimators: int
            Number of trees in the forest. Defaults to 100.
        criterion: {"squared_error", "absolute_error", "friedman_mse", "poisson"}
            The function to measure the quality of a split. "squared_error" for
            the mean squared error, "absolute_error" for the mean absolute error,
            "friedman_mse" for mean squared error with Friedman's improvement score,
            and "poisson" for Poisson deviance. Defaults to "squared_error".
        max_depth: int | None
            The maximum depth of the tree. If None, nodes are expanded until all leaves
            are pure or until all leaves contain less than min_samples_split samples.
            Defaults to None.
        min_samples_split: int
            The minimum number of samples required to split an internal node. Defaults
            to 2.
        min_samples_leaf: int
            The minimum number of samples required to be at a leaf node. Defaults to 1.
        max_features: "sqrt" | "log2" | int | float
            The number of features to consider when looking for the best split.
            If int, then consider max_features features at each split.
            If float, then max_features is a fraction and
            max(1, int(max_features * n_features_in_)) features are considered.
            If "sqrt", then max_features=sqrt(n_features).
            If "log2", then max_features=log2(n_features).
            If 1.0, then max_features=n_features.
            Defaults to 1.0.
        bootstrap: bool
            Whether bootstrap samples are used when building trees. Defaults to True.
        oob_score: bool
            Whether to use out-of-bag samples to estimate the generalization accuracy.
            Defaults to False.
        max_samples: int | None
            If bootstrap is True, the number of samples to draw from X to train each
            base estimator. If None, then draw n_samples. Defaults to None.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set. Defaults to None.
        device: DeviceLike
            Device to run the model on. If None, uses the default device. Defaults to
            "cpu".
        **kwargs: dict
            Additional keyword arguments.

        """
        _, _ = x, y  # ignore unused arguments
        TorchDeviceMixin.__init__(self, device=device, cpu_only=True)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.max_samples = max_samples
        self.random_seed = random_seed
        self.model = RandomForestRegressor(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            max_features=self.max_features,  # type: ignore reportArgumentType
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            max_samples=self.max_samples,
            random_state=self.random_seed,
        )

    @staticmethod
    def is_multioutput() -> bool:
        """Random forests support multi-output."""
        return True

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        return {
            "n_estimators": [np.random.randint(50, 500)],
            "min_samples_split": [np.random.randint(2, 20)],
            "min_samples_leaf": [np.random.randint(1, 10)],
            "max_features": ["sqrt", "log2", None, 1.0],
            "bootstrap": [True, False],
            "oob_score": [True, False],
            "max_depth": [None, *list(range(5, 30, 5))],  # None plus a range of depths
            "max_samples": [None, 0.5, 0.7, 0.9],
        }



import numpy as np
import torch
from numpy.fft import fft2, ifft2
from scipy.integrate import solve_ivp

from autoemulate.core.types import NumpyLike, TensorLike
from autoemulate.simulations.base import Simulator

integrator_keywords = {}
integrator_keywords["rtol"] = 1e-12
integrator_keywords["method"] = "RK45"
integrator_keywords["atol"] = 1e-12


class ReactionDiffusion(Simulator):
    """Simulate the reaction-diffusion PDE for a given set of parameters."""

    def __init__(  # noqa: PLR0913
        self,
        parameters_range: dict[str, tuple[float, float]] | None = None,
        output_names: list[str] | None = None,
        return_timeseries: bool = False,
        log_level: str = "progress_bar",
        n: int = 32,
        L: int = 20,
        T: float = 10.0,
        dt: float = 0.1,
    ):
        """
        Initialize the ReactionDiffusion simulator.

        Parameters
        ----------
        parameters_range: dict[str, tuple[float, float]]
            Dictionary mapping input parameter names to their (min, max) ranges.
        output_names: list[str]
            List of output parameters' names.
        log_level: str
            Logging level for the simulator. Can be one of:
            - "progress_bar": shows a progress bar during batch simulations
            - "debug": shows debug messages
            - "info": shows informational messages
            - "warning": shows warning messages
            - "error": shows error messages
            - "critical": shows critical messages
        return_timeseries: bool
            Whether to return the full timeseries or just the spatial solution at the
            final time step. Defaults to False.
        n: int
            Number of spatial points in each direction.
        L: int
            Domain size in X and Y directions.
        T: float
            Total time to simulate.
        dt: float
            Time step size.
        """
        if parameters_range is None:
            parameters_range = {"beta": (1.0, 2.0), "d": (0.05, 0.3)}
        if output_names is None:
            output_names = ["solution"]
        super().__init__(parameters_range, output_names, log_level)
        self.return_timeseries = return_timeseries
        self.n = n
        self.L = L
        self.T = T
        self.dt = dt

    def _forward(self, x: TensorLike) -> TensorLike:
        assert x.shape[0] == 1, (
            f"Simulator._forward expects a single input, got {x.shape[0]}"
        )
        u_sol, v_sol = simulate_reaction_diffusion(
            x.cpu().numpy()[0], self.return_timeseries, self.n, self.L, self.T, self.dt
        )

        # concatenate U and V arrays (flattened across time and space)
        concat_array = np.concatenate([u_sol.ravel(), v_sol.ravel()])

        # return tensor shape (1, 2*self.t*self.n*self.n)
        return torch.tensor(concat_array, dtype=torch.float32).reshape(1, -1)


def reaction_diffusion(  # noqa: PLR0913
    t: float,  # noqa: ARG001
    uvt: NumpyLike,
    K22: NumpyLike,
    d1: float,
    d2: float,
    beta: float,
    n: int,
    N: int,
):
    """
    Define the reaction-diffusion PDE in the Fourier (kx, ky) space.

    Parameters
    ----------
    t: float
        The current time step (not used).
    uvt: NumpyLike
        Fourier transformed solution vector at current time step.
    K22: NumpyLike
        The squared magnitudes of the Fourier wavevectors (kx, ky).
    d1: float
        The diffusion coefficient for species 1.
    d2: float
        The diffusion coefficient for species 2.
    beta: float
        The reaction coefficient controlling reaction between the two species.
    n: int
        Number of spatial points in each direction.
    N: int
        Total number of spatial grid points (n*n).
    """
    ut = np.reshape(uvt[:N], (n, n))
    vt = np.reshape(uvt[N : 2 * N], (n, n))
    u = np.real(ifft2(ut))
    v = np.real(ifft2(vt))
    u3 = u**3
    v3 = v**3
    u2v = (u**2) * v
    uv2 = u * (v**2)
    utrhs = np.reshape((fft2(u - u3 - uv2 + beta * u2v + beta * v3)), (N, 1))
    vtrhs = np.reshape((fft2(v - u2v - v3 - beta * u3 - beta * uv2)), (N, 1))
    uvt_reshaped = np.reshape(uvt, (len(uvt), 1))
    return np.squeeze(
        np.vstack(
            (-d1 * K22 * uvt_reshaped[:N] + utrhs, -d2 * K22 * uvt_reshaped[N:] + vtrhs)
        )
    )


def simulate_reaction_diffusion(  # noqa: PLR0913
    x: NumpyLike,
    return_timeseries: bool = False,
    n: int = 32,
    L: int = 20,
    T: float = 10.0,
    dt: float = 0.1,
) -> tuple[NumpyLike, NumpyLike]:
    """
    Simulate the reaction-diffusion PDE for a given set of parameters.

    Parameters
    ----------
    x: NumpyLike
        The parameters of the reaction-diffusion model. The first element is the
        reaction coefficient (beta) and the second element is the diffusion
        coefficient (d).
    return_timeseries: bool
        Whether to return the full timeseries or just the spatial solution at the final
        time step. Defaults to False.
    n: int
        Number of spatial points in each direction. Defaults to 32.
    L: int
        Domain size in X and Y directions. Defaults to 20.
    T: float
        Total time to simulate. Defaults to 10.0.
    dt: float
        Time step size. Defaults to 0.1.

    Returns
    -------
    tuple[NumpyLike, NumpyLike]
        [u_sol, v_sol], the spatial solution of the reaction-diffusion PDE, either as a
        timeseries or at the final time point of `return_timeseries` is False.
    """
    beta, d = x
    d1 = d2 = d

    t = np.linspace(0, T, int(T / dt))

    N = n * n
    x_uniform = np.linspace(-L / 2, L / 2, n + 1)
    x_grid = x_uniform[:n]
    y_grid = x_uniform[:n]
    n2 = int(n / 2)
    # Define Fourier wavevectors (kx, ky)
    kx = (2 * np.pi / L) * np.hstack(
        (np.linspace(0, n2 - 1, n2), np.linspace(-n2, -1, n2))
    )
    ky = kx
    # Get 2D meshes in (x, y) and (kx, ky)
    X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
    KX, KY = np.meshgrid(kx, ky)
    K2 = KX**2 + KY**2
    K22 = np.reshape(K2, (N, 1))

    m = 1

    # define our solution vectors
    u = np.zeros((len(x_grid), len(y_grid), len(t)))
    v = np.zeros((len(x_grid), len(y_grid), len(t)))

    # Initial conditions
    u[:, :, 0] = np.tanh(np.sqrt(X_grid**2 + Y_grid**2)) * np.cos(
        m * np.angle(X_grid + 1j * Y_grid) - (np.sqrt(X_grid**2 + Y_grid**2))
    )
    v[:, :, 0] = np.tanh(np.sqrt(X_grid**2 + Y_grid**2)) * np.sin(
        m * np.angle(X_grid + 1j * Y_grid) - (np.sqrt(X_grid**2 + Y_grid**2))
    )

    # uvt is the solution vector in Fourier space, so below
    # we are initializing the 2D FFT of the initial condition, uvt0
    uvt0 = np.squeeze(
        np.hstack(
            (np.reshape(fft2(u[:, :, 0]), (1, N)), np.reshape(fft2(v[:, :, 0]), (1, N)))
        )
    )

    # Solve the PDE in the Fourier space, where it reduces to system of ODEs
    uvsol = solve_ivp(
        reaction_diffusion,
        (t[0], t[-1]),
        y0=uvt0,
        t_eval=t,
        args=(K22, d1, d2, beta, n, N),
        **integrator_keywords,
    )
    uvsol = uvsol.y

    # Reshape things and ifft back into (x, y, t) space from (kx, ky, t) space
    for j in range(len(t)):
        ut = np.reshape(uvsol[:N, j], (n, n))
        vt = np.reshape(uvsol[N:, j], (n, n))
        u[:, :, j] = np.real(ifft2(ut))
        v[:, :, j] = np.real(ifft2(vt))

    if return_timeseries:
        return u.transpose(2, 0, 1), v.transpose(2, 0, 1)
    # Return the last snapshot
    u_sol = u[:, :, -1]
    v_sol = v[:, :, -1]
    return u_sol, v_sol



from abc import abstractmethod
from dataclasses import dataclass

from ..core.types import GaussianLike, TensorLike
from .base import Active


@dataclass(kw_only=True)
class Membership(Active):
    """
    Active learning strategy based on membership queries.

    Parameters
    ----------
    (Inherits parameters from Active)
    """

    @abstractmethod
    def query(
        self, x: TensorLike | None = None
    ) -> tuple[TensorLike | None, TensorLike | GaussianLike, dict[str, float]]:
        """
        Abstract method to query new samples using a membership strategy.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor, Dict[str, List[Any]]]
            A tuple containing:
            - The queried samples,
            - The predicted outputs,
            - The covariance estimates,
            - A dictionary of additional metrics.
        """



from copy import deepcopy

import torch
from torch import nn


class EarlyStoppingException(Exception):
    """Custom exception to signal early stopping during training."""

    def __init__(
        self, message: str = "Training stopped early due to early stopping criteria."
    ):
        super().__init__(message)


class EarlyStopping:
    """
    Early stopping callback for PyTorch models.

    Stop training early if the training loss did not improve in `patience` number of
    epochs by at least `threshold` value. Can be used inside the training loop of any
    PyTorch model.

    Parameters
    ----------
    lower_is_better: bool
        Whether lower scores should be considered an improvement. Defaults to True.
    patience: int
        Number of epochs to wait for improvement of the training loss until the
        training process is stopped. Defaults to 5.
    threshold: int
        Minimum change in training loss (i.e., `min_delta`) that is considered an
        improvement in performance. Defaults to 1e-4.
    threshold_mode: str
        One of `rel`, `abs`. Decides whether the `threshold` value is interpreted in
        absolute terms or as a fraction of the best  score so far (relative). Defaults
        to `rel`.
    load_best: bool
        Whether to restore module weights from the epoch with the best value of
        training loss. If False, the module weights obtained at the last step of
        training are used. Defaults to False.

    Notes
    -----
    This class is practically identical to `EarlyStopping` in skorch. The main
    difference is that the method `_calc_new_threshold`, is corrected to ensure
    monotonicity. We also do not have the option to monitor other metrics (e.g.,
    validation loss) instead of the train loss.
    """

    def __init__(
        self,
        patience: int = 5,
        threshold: float = 1e-4,
        threshold_mode: str = "rel",
        lower_is_better: bool = True,
        load_best: bool = False,
    ):
        self.lower_is_better = lower_is_better
        self.patience = patience
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.misses_ = 0
        self.dynamic_threshold_ = None
        self.load_best = load_best

    def __getstate__(self):
        """Return state without pickling the best model weights."""
        # Avoids having to save the module_ weights twice when pickling the model
        state = self.__dict__.copy()
        state["best_model_weights_"] = None
        return state

    def on_train_begin(self):
        """Initialize early stopping parameters at the start of training."""
        if self.threshold_mode not in ["rel", "abs"]:
            raise ValueError(f"Invalid threshold mode: '{self.threshold_mode}'")
        self.misses_ = 0
        self.dynamic_threshold_ = torch.inf if self.lower_is_better else -torch.inf
        self.best_model_weights_ = None
        self.best_epoch_ = 0

    def on_epoch_end(self, model: nn.Module, curr_epoch: int, curr_score: float):
        """
        Determine whether to stop training at this epoch and save best model weights.

        Parameters
        ----------
        model: nn.Module
            A PyTorch model.
        curr_epoch: int
            The current epoch.
        curr_score: float
            The current training loss.
        """
        if not self._is_score_improved(curr_score):
            self.misses_ += 1
        else:
            self.misses_ = 0
            self.dynamic_threshold_ = self._calc_new_threshold(curr_score)
            self.best_epoch_ = curr_epoch
            if self.load_best:
                self.best_model_weights_ = deepcopy(model.state_dict())
        if self.misses_ == self.patience:
            msg = (
                "Stopping since train loss has not improved in the last "
                f"{self.patience} epochs."
            )
            raise EarlyStoppingException(msg)

    def on_train_end(self, model: nn.Module, last_epoch: int):
        """
        Optionally restore module weights from the epoch with the best train loss.

        Parameters
        ----------
        model: nn.Module
            A PyTorch model.
        last_epoch: int
            The number of completed epochs before training was stopped.
        """
        if (
            self.load_best
            and (self.best_epoch_ != last_epoch)
            and (self.best_model_weights_ is not None)
        ):
            model.load_state_dict(self.best_model_weights_)

    def _is_score_improved(self, score):
        if self.lower_is_better:
            return score < self.dynamic_threshold_
        return score > self.dynamic_threshold_

    def _calc_new_threshold(self, score):
        """
        Determine threshold based on score.

        Notes
        -----
        This function updates the skorch version, which assumes that the score returned
        by the cost function is always positive.
        PR to skorch pending: https://github.com/skorch-dev/skorch/pull/1065
        """
        if self.threshold_mode == "rel":
            abs_threshold_change = self.threshold * abs(score)
        else:
            abs_threshold_change = self.threshold

        if self.lower_is_better:
            new_threshold = score - abs_threshold_change
        else:
            new_threshold = score + abs_threshold_change
        return new_threshold



import gpytorch
import torch

from autoemulate.core.types import TensorLike

from .polynomial_features import PolynomialFeatures


class PolyMean(gpytorch.means.Mean):
    """
    A general polynomial mean module for Gaussian Process emulators.

    Parameters
    ----------
    degree: int
        The degree of the polynomial for which we are defining
        the mapping.
    input_size: int
        The number of features to be mapped.
    batch_shape: int | None
        Optional batch dimension.
    bias: bool
        Flag for including a bias in the definition of the polynomial.
        If set to `False` polynomial includes weights only.
    """

    def __init__(
        self,
        degree: int,
        input_size: int,
        batch_shape: torch.Size | None = None,
        bias=True,
    ):
        super().__init__()
        self.degree = degree
        self.input_size = input_size

        if batch_shape is None:
            batch_shape = torch.Size()

        self.poly = PolynomialFeatures(self.degree, self.input_size)
        self.poly.fit()

        assert self.poly.indices is not None
        n_weights = len(self.poly.indices)
        self.register_parameter(
            name="weights",
            parameter=torch.nn.Parameter(torch.randn(*batch_shape, n_weights, 1)),
        )

        if bias:
            self.register_parameter(
                name="bias", parameter=torch.nn.Parameter(torch.randn(*batch_shape, 1))
            )
        else:
            self.bias = None

    def forward(self, x: TensorLike):
        """Forward pass through the polynomial mean module."""
        x_ = self.poly.transform(x)
        assert isinstance(self.weights, TensorLike)
        res = x_.matmul(self.weights).squeeze(-1)
        if self.bias is not None:
            res = res + self.bias
        return res

    def __repr__(self):
        """Return the string representation of the PolyMean module."""
        return f"PolyMean(degree={self.degree}, input_size={self.input_size})"



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from inspect import isabstract

import torch
from anytree import Node, RenderTree
from torch.distributions import MultivariateNormal
from torcheval.metrics import MeanSquaredError, R2Score

from autoemulate.core.logging_config import get_configured_logger
from autoemulate.data.utils import ValidationMixin
from autoemulate.emulators.base import Emulator
from autoemulate.simulations.base import Simulator

from ..core.types import GaussianLike, TensorLike


@dataclass(kw_only=True)
class Learner(ValidationMixin, ABC):
    """
    Learner class that combines a simulator and an emulator for active learning.

    The learner uses a simulator to generate ground-truth outputs and an emulator to
    approximate the simulator. Training data is stored and used to update the emulator.

    Parameters
    ----------
    simulator: Simulator
        Simulator instance used to generate ground truth outputs.
    emulator: Emulator
        Emulator instance used to approximate the simulator.
    x_train: TensorLike
        Initial training input tensor.
    y_train: TensorLike
        Initial training output tensor.
    """

    simulator: Simulator
    emulator: Emulator
    x_train: TensorLike
    y_train: TensorLike
    log_level: str = "progress_bar"
    in_dim: int = field(init=False)
    out_dim: int = field(init=False)

    def __post_init__(self):
        """Initialize the learner with training data and fit the emulator."""
        log_level = getattr(self, "log_level", "progress_bar")
        self.logger, self.progress_bar = get_configured_logger(log_level)
        self.logger.info("Initializing Learner with training data.")
        self.emulator.fit(self.x_train, self.y_train)
        self.logger.info("Emulator fitted with initial training data.")
        self.in_dim = self.x_train.shape[1]
        self.out_dim = self.y_train.shape[1]

    @classmethod
    def registry(cls) -> dict:
        """
        Build a registry of all subclasses.

        Recursively builds a dictionary mapping class names to class objects,
        starting from cls.

        Returns
        -------
        dict
            A dictionary where keys are class names and values are the class objects.
        """
        d = {cls.__name__: cls}
        for sub in cls.__subclasses__():
            d.update(sub.registry())
        return d

    @classmethod
    def hierarchy(cls) -> list[tuple[str, str]]:
        """
        Build a list of inheritance relationships.

        Recursively collects the inheritance relationships (as ordered pairs)
        starting from cls.

        Returns
        -------
        list[tuple[str, str]]
            A list of tuples where each tuple is (parent_class_name, child_class_name).
        """
        edges = []
        for base in cls.__subclasses__():
            edges.append((cls.__name__, base.__name__))
            edges.extend(base.hierarchy())
        return edges

    @classmethod
    def plot_hierarchy(cls):
        """
        Print a tree representation of the class hierarchy.

        Builds and prints an ASCII tree of the class hierarchy starting from cls.

        Each class name is annotated with a marker indicating whether it is abstract:
          - [Abstract] for classes with one or more abstract methods.
          - [Concrete] for classes that are fully implemented.

        The method uses the anytree library to construct and render the tree.
        """
        pairs = cls.hierarchy()
        nodes = {name: Node(name) for pair in pairs for name in pair}
        for parent, child in pairs:
            nodes[child].parent = nodes[parent]
        root = nodes[cls.__name__]
        registry = cls.registry()
        for pre, _, node in RenderTree(root):
            class_obj = registry.get(node.name)
            mark = "[Abstract]" if class_obj and isabstract(class_obj) else "[Concrete]"
            print(f"{pre}{node.name} {mark}")


@dataclass(kw_only=True)
class Active(Learner):
    """Active learning learner that combines a simulator and an emulator."""

    def __post_init__(self):  # noqa: D105
        super().__post_init__()
        self.metrics = {
            k: []
            for k in ["mse", "r2", "rate", "logdet", "trace", "max_eigval", "n_queries"]
        }
        self.mse = MeanSquaredError()
        self.r2 = R2Score()
        self.n_queries = 0

    def fit(self, *args):
        """
        Fit the emulator with new training data and update metrics.

        This method queries the simulator for new data, appends it to the training set,
        and refits the emulator. It also updates various metrics based on the new data.
        """
        # Query simulator and fit emulator
        x, output, extra = self.query(*args)
        if isinstance(output, TensorLike):
            y_pred = output
        elif isinstance(output, GaussianLike):
            assert output.variance.ndim == 2
            y_pred, _ = output.mean, output.variance
        elif isinstance(output, GaussianLike):
            y_pred, _ = output.loc, None
        else:
            msg = (
                f"Output must be either `Tensor` or `MultivariateNormal` but got "
                f"{type(output)}"
            )
            raise TypeError(msg)

        if x is not None:
            # If x is not, we skip the point (typically for Stream learners)
            self.logger.info("Appending new training data and refitting emulator.")
            y_true = self.simulator.forward(x)
            assert isinstance(y_true, TensorLike)
            self.x_train = torch.cat([self.x_train, x])
            self.y_train = torch.cat([self.y_train, y_true])
            self.emulator.fit(self.x_train, self.y_train)
            self.mse.update(y_pred, y_true)
            self.r2.update(y_pred, y_true)
            self.n_queries += 1
            self.logger.info("Training data updated. Total queries: %s", self.n_queries)

        # Only compute once we have ≥2 labeled points
        if self.n_queries >= 2:
            mse_val = self.mse.compute().item()
            r2_val = self.r2.compute().item()
        else:
            mse_val = float("nan")
            r2_val = float("nan")

        self.metrics["mse"].append(mse_val)
        self.metrics["r2"].append(r2_val)
        self.metrics["rate"].append(self.n_queries / (len(self.metrics["rate"]) + 1))
        self.metrics["n_queries"].append(self.n_queries)
        self.logger.info("Metrics updated: MSE=%s, R2=%s", mse_val, r2_val)

        # If Gaussian output
        # TODO: check generality for other GPs (e.g. with full covariance)
        if isinstance(output, MultivariateNormal):
            assert isinstance(output.variance, TensorLike)
            assert output.variance.ndim == 2
            assert output.variance.shape[1] == self.out_dim
            # For Multivariate Normal, the variance property gives the correct value
            # required here with shape: (batch, out_dim)
            covariance = output.variance
            self.metrics["trace"].append(self.trace(covariance, self.out_dim).item())
            self.metrics["logdet"].append(self.logdet(covariance, self.out_dim).item())
            self.metrics["max_eigval"].append(self.max_eigval(covariance).item())
            self.logger.info("Gaussian output metrics updated.")

        # extra per-strategy metrics
        for k, v in extra.items():
            self.metrics.setdefault(k, []).append(v)
            self.logger.info("Extra metric '%s' updated: %s", k, v)

    @property
    def summary(self):
        """
        Summary metrics for the active learner.

        Compute summary metrics for the active learner based on recorded learning
        histories.

        This property converts the history of MSE and cumulative query counts into float
        tensors, and then computes two types of summary metrics:

        1. **Per-Query Ratios:** For each metric in ("mse", "r2", "trace", "logdet",
           "max_eigval"), the ratio is defined as the last recorded value of the metric
           divided by the last recorded cumulative number of queries. If no queries have
           been made (i.e. the last query count is 0), NaN is returned for all per-query
           ratios.

        2. **Area Under the MSE Curve (AUC):** The area is computed via trapezoidal
           integration of the MSE values over the cumulative number of queries, but only
           if there are at least 2 valid MSE entries. Otherwise, the AUC is set to NaN.

        Returns
        -------
        dict
            A dictionary containing:
                - "<metric>_per_query": The per-query ratio for each metric in ("mse",
                  "r2", "trace", "logdet", "max_eigval").
                - "auc_mse": The area under the MSE curve computed with respect to the
                  cumulative number of queries.
        """
        # pull histories into float tensors
        mse = torch.tensor(self.metrics["mse"], dtype=torch.float32)
        q = torch.tensor(self.metrics["n_queries"], dtype=torch.float32)

        # build per-query ratios safely (avoid zero division)
        d = {}
        if q[-1] > 0:
            for k in ("mse", "r2", "trace", "logdet", "max_eigval"):
                d[f"{k}_per_query"] = (
                    self.metrics[k][-1] / self.metrics["n_queries"][-1]
                )
        else:
            for k in ("mse", "r2", "trace", "logdet", "max_eigval"):
                d[f"{k}_per_query"] = float("nan")

        # mask out nan entries for MSE
        valid_mse = ~torch.isnan(mse)

        # compute AUC only if ≥2 valid points
        d["auc_mse"] = (
            torch.trapz(mse[valid_mse], q[valid_mse]).item()
            if valid_mse.sum() >= 2
            else float("nan")
        )
        return d

    @abstractmethod
    def query(
        self, x: TensorLike | None = None
    ) -> tuple[TensorLike | None, TensorLike | GaussianLike, dict[str, float]]:
        """
        Abstract method to query new samples.

        Parameters
        ----------
        *arg: TensorLike or None
            Optional input samples.

        Returns
        -------
        tuple[TensorLike or None, TensorLike, TensorLike, Dict[str, list[Any]]]
            A tuple containing:
            - The queried samples (or None if no query is made),
            - The predicted outputs,
            - The covariance estimates,
            - A dictionary of additional metrics.
        """



import arviz as az
import pyro
import pyro.distributions as dist
import torch
from pyro.infer import HMC, MCMC, NUTS, Predictive
from pyro.infer.mcmc import RandomWalkKernel

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.logging_config import get_configured_logger
from autoemulate.core.types import DeviceLike, DistributionLike, TensorLike
from autoemulate.emulators.base import Emulator


class BayesianCalibration(TorchDeviceMixin):
    """
    Bayesian calibration using Markov Chain Monte Carlo (MCMC).

    Bayesian calibration estimates the probability distribution over input parameters
    given observed data, providing uncertainty estimates.
    """

    def __init__(  # noqa: PLR0913
        self,
        emulator: Emulator,
        parameter_range: dict[str, tuple[float, float]],
        observations: dict[str, TensorLike],
        observation_noise: float | dict[str, float] = 0.1,
        calibration_params: list[str] | None = None,
        device: DeviceLike | None = None,
        log_level: str = "progress_bar",
    ):
        """
        Initialize the HMC calibration object.

        Parameters
        ----------
        emulator: Emulator
            Fitted Emulator object.
        parameters_range : dict[str, tuple[float, float]]
            A dictionary mapping input parameter names to their (min, max) ranges.
        observations: dict[str, TensorLike]
            A dictionary of observations for each output.
        observation_noise: float |  dict[str, float]
            A single value or a dictionary of values (one per output). Defaults to 0.1.
        calibration_params: list[str] | None
            Optional list of input parameters to calibrate. Any parameters that are not
            listed will be set to the midpoint value of their parameter range. If None,
            will calibrate all input parameters. Defaults to None.
        device: DeviceLike | None
            The device to use. If None, the default torch device is returned.
            TODO: do we need to do anything more to ensure the device is correctly
            handled for the pyro model?
        log_level: str
            Logging level for the calibration. Can be one of:
            - "progress_bar": shows a progress bar during batch simulations
            - "debug": shows debug messages
            - "info": shows informational messages
            - "warning": shows warning messages
            - "error": shows error messages
            - "critical": shows critical messages

        Notes
        -----
        The model assumes:
        - Uniform priors for calibrated parameters (bounds given by `parameters_range`)
        - Gaussian likelihood with no correlation between outputs
        All non-calibrated parameters are set to a constant value. This is chosen as the
        midpoint value of `parameters_range`.
        """
        TorchDeviceMixin.__init__(self, device=device)
        self.parameter_range = parameter_range
        if calibration_params is None:
            calibration_params = list(parameter_range.keys())
        self.calibration_params = calibration_params
        self.emulator = emulator
        self.emulator.device = self.device
        self.output_names = list(observations.keys())
        self.logger, self.progress_bar = get_configured_logger(log_level)
        self.logger.info(
            "Initializing BayesianCalibration with parameters: %s",
            self.calibration_params,
        )

        # Check observation tensors are 1D (convert if 0D)
        processed_observations = {}
        obs_lengths = []
        for output, obs in observations.items():
            if obs.ndim == 0:
                corrected_obs = obs.unsqueeze(0)
                self.logger.debug(
                    "Observation for output '%s' converted from 0D to 1D.",
                    output,
                )
            elif obs.ndim > 1:
                self.logger.error("Tensor for output '%s' is not 1D.", output)
                raise ValueError(f"Tensor for output '{output}' is not 1D.")
            else:
                corrected_obs = obs
            processed_observations[output] = corrected_obs.to(self.device)
            obs_lengths.append(corrected_obs.shape[0])
        if len(set(obs_lengths)) != 1:
            msg = "All outputs must have the same number of observations."
            self.logger.error(msg)
            raise ValueError(msg)
        self.observations = processed_observations
        self.n_observations = obs_lengths[0]
        self.logger.info("Processed observations for outputs: %s", self.output_names)

        # Save observation noise as {output: value} dictionary
        if isinstance(observation_noise, float):
            self.observation_noise = dict.fromkeys(
                self.output_names,
                torch.tensor(observation_noise).to(self.device),
            )
            self.logger.debug("Observation noise set as float: %s", observation_noise)
        elif isinstance(observation_noise, dict):
            self.observation_noise = observation_noise
            self.logger.debug("Observation noise set as dict: %s", observation_noise)
        else:
            msg = "Noise must be either a float or a dictionary of floats."
            self.logger.error(msg)
            raise ValueError(msg)

    def _get_kernel(self, sampler: str, **sampler_kwargs):
        """Get the appropriate MCMC kernel based on sampler choice."""
        sampler = sampler.lower()

        if sampler == "nuts":
            self.logger.debug("Using NUTS kernel.")
            return NUTS(self.model, **sampler_kwargs)
        if sampler == "hmc":
            step_size = sampler_kwargs.pop("step_size", 0.01)
            trajectory_length = sampler_kwargs.pop("trajectory_length", 1.0)
            self.logger.debug(
                "Using HMC kernel with step_size=%s, trajectory_length=%s",
                step_size,
                trajectory_length,
            )
            return HMC(
                self.model,
                step_size=step_size,
                trajectory_length=trajectory_length,
                **sampler_kwargs,
            )
        if sampler == "metropolis":
            self.logger.debug("Using Metropolis (RandomWalkKernel).")
            return RandomWalkKernel(self.model, **sampler_kwargs)
        self.logger.error("Unknown sampler: %s", sampler)
        raise ValueError(f"Unknown sampler: {sampler}")

    def model(self, predict: bool = False):
        """
        Pyro model.

        Parameters
        ----------
        predict: bool
            Whether to run the model with existing samples to generate posterior
            predictive distribution. Used with `pyro.infer.Predictive`.
        """
        # Pre-allocate tensor for all input parameters, shape [1, n_inputs]
        param_list = []
        # Each param is either sampled (if calibrated) or set to a constant value
        for param in self.parameter_range:
            if param in self.calibration_params:
                # Sample from uniform prior
                min_val, max_val = self.parameter_range[param]
                sampled_val = pyro.sample(param, dist.Uniform(min_val, max_val))
                param_list.append(sampled_val.to(self.device))
            else:
                # Set to midpoint value in parameter range
                min_val, max_val = self.parameter_range[param]
                midpoint_val = (min_val + max_val) / 2
                param_list.append(torch.tensor(midpoint_val, device=self.device))
        full_params = torch.stack(param_list, dim=0).unsqueeze(0).float()

        # Get emulator prediction
        output = self.emulator.predict(full_params, with_grad=True)
        if isinstance(output, TensorLike):
            pred_mean = output.to(self.device)
        elif isinstance(output, DistributionLike):
            pred_mean = output.mean.to(self.device)
        else:
            msg = "The emulator did not return a tensor or a distribution object."
            self.logger.error(msg)
            raise ValueError(msg)

        # Likelihood
        for i, output in enumerate(self.output_names):
            if not predict:
                with pyro.plate(f"data_{output}", self.n_observations):
                    pyro.sample(
                        output,
                        dist.Normal(pred_mean[0, i], self.observation_noise[output]),
                        obs=self.observations[output],
                    )
            else:
                with pyro.plate(f"data_{output}", self.n_observations):
                    pyro.sample(
                        output,
                        dist.Normal(pred_mean[0, i], self.observation_noise[output]),
                    )

    def run_mcmc(
        self,
        warmup_steps: int = 500,
        num_samples: int = 1000,
        num_chains: int = 1,
        initial_params: dict[str, TensorLike] | None = None,
        sampler: str = "nuts",
        **sampler_kwargs,
    ) -> MCMC:
        """
        Run Markov Chain Monte Carlo (MCMC). Defaults to using the NUTS sampler.

        Parameters
        ----------
        warmup_steps: int
            Number of warm up steps to run per chain (i.e., burn-in). These samples are
            discarded. Defaults to 500.
        num_samples: int
            Number of samples to draw after warm up. Defaults to 1000.
        num_chains: int
            Number of parallel chains to run. Defaults to 1.
        initial_params: dict[str, TensorLike] | None
            Optional dictionary specifiying initial values for each calibration
            parameter per chain. The tensors must be of length `num_chains`.
        sampler: str
            The MCMC kernel to use, one of "hmc", "nuts" or "metropolis".

        Returns
        -------
        MCMC
            The Pyro MCMC object. Methods include `summary()` and `get_samples()`.
        """
        # Check initial param values match number of chains

        if initial_params is not None:
            for param, init_vals in initial_params.items():
                if init_vals.shape[0] != num_chains:
                    msg = (
                        "An initial value must be provided for each chain, parameter "
                        f"{param} tensor only has {init_vals.shape[0]} values."
                    )
                    self.logger.error(msg)
                    raise ValueError(msg)
            self.logger.debug(
                "Initial parameters provided for MCMC: %s", initial_params
            )

        # Run NUTS
        kernel = self._get_kernel(sampler, **sampler_kwargs)
        mcmc = MCMC(
            kernel,
            warmup_steps=warmup_steps,
            num_samples=num_samples,
            num_chains=num_chains,
            # If None, init values are sampled from the prior.
            initial_params=initial_params,
            # Multiprocessing
            mp_context="spawn" if num_chains > 1 else None,
        )
        self.logger.info("Starting MCMC run.")
        mcmc.run()
        self.logger.info("MCMC run completed.")
        return mcmc

    def posterior_predictive(self, mcmc: MCMC) -> TensorLike:
        """
        Return posterior predictive samples.

        Parameters
        ----------
        mcmc: MCMC
            The MCMC object.

        Returns
        -------
        TensorLike
            Tensor of posterior predictive samples [n_mcmc_samples, n_obs, n_outputs].
        """
        posterior_samples = mcmc.get_samples()
        posterior_predictive = Predictive(self.model, posterior_samples)
        samples = posterior_predictive(predict=True)
        self.logger.debug("Posterior predictive samples generated.")
        return samples

    def to_arviz(
        self, mcmc: MCMC, posterior_predictive: bool = False
    ) -> az.InferenceData:
        """
        Convert MCMC object to Arviz InferenceData object for plotting.

        Parameters
        ----------
        mcmc: MCMC
            The MCMC object.
        posterior_predictive: bool
            Whether to include posterior predictive samples. Defaults to False.

        Returns
        -------
        az.InferenceData
        """
        pp_samples = None
        if posterior_predictive:
            self.logger.info("Including posterior predictive samples in Arviz output.")
            pp_samples = self.posterior_predictive(mcmc)

        # Need to create dataset manually for Metropolis Hastings
        # This is because az.from_pyro expects kernel with `divergences`
        if isinstance(mcmc.kernel, RandomWalkKernel):
            self.logger.debug(
                "Using manual conversion for Metropolis (RandomWalkKernel) kernel."
            )
            if posterior_predictive:
                az_data = az.InferenceData(
                    posterior=az.convert_to_dataset(
                        mcmc.get_samples(group_by_chain=True)
                    ),
                    posterior_predictive=az.convert_to_dataset(pp_samples),
                    observed_data=az.convert_to_dataset(self.observations),
                )
            else:
                az_data = az.InferenceData(
                    posterior=az.convert_to_dataset(
                        mcmc.get_samples(group_by_chain=True)
                    ),
                )
        else:
            self.logger.debug("Using az.from_pyro for conversion.")
            az_data = az.from_pyro(mcmc, posterior_predictive=pp_samples)

        self.logger.info("Arviz InferenceData conversion complete.")
        return az_data



import torch
from gpytorch.means import ConstantMean, LinearMean, ZeroMean

from .poly_mean import PolyMean


def constant_mean(n_features: int | None, n_outputs: torch.Size | None) -> ConstantMean:
    """
    ConstantMean module.

    Parameters
    ----------
    n_features: int | None
        Number of input features. This parameter is not used in the ConstantMean module.
    n_outputs: torch.Size | None
        Batch shape of the mean. If None, the mean is not initialized with a batch
        shape.

    Returns
    -------
    ConstantMean
        The initialized ConstantMean module.
    """
    _ = n_features  # Unused parameter
    return (
        ConstantMean(batch_shape=n_outputs) if n_outputs is not None else ConstantMean()
    )


def zero_mean(n_features: int, n_outputs: torch.Size | None) -> ZeroMean:
    """ZeroMean module.

    Parameters
    ----------
    n_features: int
        Number of input features. This parameter is not used in the ZeroMean module.
    n_outputs: torch.Size | None
        Batch shape of the mean. If None, the mean is not initialized with a batch
        shape.

    Returns
    -------
    ZeroMean
        The initialized ZeroMean module.
    """
    _ = n_features  # Unused parameter
    return ZeroMean(batch_shape=n_outputs) if n_outputs is not None else ZeroMean()


def linear_mean(n_features: int, n_outputs: torch.Size | None) -> LinearMean:
    """
    LinearMean module.

    Parameters
    ----------
    n_features: int
        Number of input features.
    n_outputs: torch.Size | None
        Batch shape of the mean. If None, the mean is not initialized with a batch
        shape.

    Returns
    -------
    LinearMean
        The initialized LinearMean module.
    """
    return (
        LinearMean(input_size=n_features, batch_shape=n_outputs)
        if n_outputs is not None
        else LinearMean(input_size=n_features)
    )


def poly_mean(n_features: int, n_outputs: torch.Size | None) -> PolyMean:
    """
    PolyMean module (quadratic polynomial mean).

    Parameters
    ----------
    n_features: int
        Number of input features.
    n_outputs: torch.Size | None
        Batch shape of the mean. If None, the mean is not initialized with a batch
        shape.

    Returns
    -------
    PolyMean
        The initialized PolyMean module.
    """
    return (
        PolyMean(degree=2, input_size=n_features, batch_shape=n_outputs)
        if n_outputs is not None
        else PolyMean(degree=2, input_size=n_features)
    )



from .base import Emulator
from .ensemble import EnsembleMLP, EnsembleMLPDropout
from .gaussian_process.exact import GaussianProcess, GaussianProcessCorrelated
from .lightgbm import LightGBM
from .nn.mlp import MLP
from .radial_basis_functions import RadialBasisFunctions
from .random_forest import RandomForest
from .svm import SupportVectorMachine
from .transformed.base import TransformedEmulator

# from .neural_processes.conditional_neural_process import CNPModule

ALL_EMULATORS: list[type[Emulator]] = [
    GaussianProcess,
    GaussianProcessCorrelated,
    LightGBM,
    # CNPModule,
    SupportVectorMachine,
    RadialBasisFunctions,
    RandomForest,
    MLP,
    EnsembleMLP,
    EnsembleMLPDropout,
]

EMULATOR_REGISTRY = {em_cls.model_name().lower(): em_cls for em_cls in ALL_EMULATORS}
EMULATOR_REGISTRY_SHORT_NAME = {em_cls.short_name(): em_cls for em_cls in ALL_EMULATORS}


def get_emulator_class(name: str) -> type[Emulator]:
    """
    Get the emulator class by name.

    Parameters
    ----------
    name: str
        The name of the emulator class.

    Returns
    -------
    type[Emulator] | None
        The emulator class if found, None otherwise.
    """
    emulator_cls = EMULATOR_REGISTRY.get(
        name.lower()
    ) or EMULATOR_REGISTRY_SHORT_NAME.get(name.lower())

    if emulator_cls is None:
        raise ValueError(
            f"Unknown emulator name: {name}.Available: {list(EMULATOR_REGISTRY.keys())}"
        )

    return emulator_cls


__all__ = [
    "MLP",
    "EnsembleMLP",
    "EnsembleMLPDropout",
    "GaussianProcess",
    "GaussianProcessCorrelated",
    "LightGBM",
    "RadialBasisFunctions",
    "RandomForest",
    "SupportVectorMachine",
    "TransformedEmulator",
]



import random

import numpy as np
import torch
from autoemulate.core.types import (
    InputLike,
    OutputLike,
    TensorLike,
    TorchScalarDType,
)
from sklearn.utils.validation import check_X_y
from torch.utils.data import DataLoader, Dataset, Subset, TensorDataset, random_split


class ConversionMixin:
    """Mixin class to convert input data to PyTorch Datasets and DataLoaders."""

    @classmethod
    def _convert_to_dataset(
        cls,
        x: InputLike,
        y: InputLike | None = None,
    ) -> Dataset:
        """Convert input data to PyTorch Dataset."""
        # Convert input to Dataset if not already
        if isinstance(x, np.ndarray):
            x = torch.tensor(x, dtype=torch.float32)
        if isinstance(y, np.ndarray):
            y = torch.tensor(y, dtype=torch.float32)

        if isinstance(x, torch.Tensor | np.ndarray) and isinstance(
            y, torch.Tensor | np.ndarray
        ):
            dataset = TensorDataset(x, y)
        elif isinstance(x, torch.Tensor | np.ndarray) and y is None:
            dataset = TensorDataset(x)
        elif isinstance(x, Dataset) and y is None:
            dataset = x
        elif isinstance(x, DataLoader) and y is None:
            dataset = x.dataset
        else:
            raise ValueError(
                f"Unsupported type for x ({type(x)}). Must be numpy array or PyTorch "
                "tensor."
            )

        return dataset

    @classmethod
    def _convert_to_dataloader(
        cls,
        x: InputLike,
        y: InputLike | None = None,
        batch_size: int = 16,
        shuffle: bool = True,
    ) -> DataLoader:
        """Convert input data to PyTorch DataLoaders."""
        if isinstance(x, DataLoader) and y is None:
            dataloader = x
        elif isinstance(x, DataLoader) and y is not None:
            raise ValueError(
                f"Since x is already a DataLoader, expect y to be None, not {type(y)}."
            )
        else:
            dataset = cls._convert_to_dataset(x, y)
            dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        return dataloader

    @classmethod
    def _convert_to_tensors(
        cls,
        x: InputLike,
        y: InputLike | None = None,
        dtype: torch.dtype = torch.float32,
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        """Convert InputLike x, y to Tensor or tuple of Tensors."""
        dataset = cls._convert_to_dataset(x, y)

        # Handle Subset of TensorDataset
        if isinstance(dataset, Subset):
            ds = dataset
            while isinstance(ds, Subset):
                ds = ds.dataset
            if isinstance(ds, TensorDataset):
                tensors = ds.tensors
                indices = dataset.indices

                # Use indexing to get subset tensors
                subset_tensors = tuple(tensor[indices] for tensor in tensors)
                dataset = TensorDataset(*subset_tensors)
            else:
                raise ValueError(
                    f"Subset must wrap a TensorDataset. Found {type(dataset.dataset)}."
                )

        if isinstance(dataset, TensorDataset):
            if len(dataset.tensors) > 2:
                raise ValueError(
                    f"Dataset must have 2 or fewer tensors. Found "
                    f"{len(dataset.tensors)}."
                )
            if len(dataset.tensors) == 2:
                x, y = dataset.tensors
                assert x.ndim == 2
                assert y.ndim in (1, 2)
                # Ensure always 2D tensors
                if y.ndim == 1:
                    y = y.unsqueeze(1)
                return x.to(dtype), y.to(dtype)
            if len(dataset.tensors) == 1:
                (x,) = dataset.tensors
                assert x.ndim == 2
                return x.to(dtype)
            msg = "Number of tensors returned must be greater than zero."
            raise ValueError(msg)
        # Note: this error will never be raised, the same error is raised in
        # _convert_to_dataset
        raise ValueError(
            f"Unsupported type for x ({type(x)}). Must be numpy array or PyTorch "
            "tensor."
        )

    def _convert_to_numpy(
        self,
        x: InputLike,
        y: InputLike | None = None,
    ) -> tuple[np.ndarray, np.ndarray | None]:
        """Convert InputLike x, y to tuple of numpy arrays."""
        if isinstance(x, np.ndarray) and (y is None or isinstance(y, np.ndarray)):
            return x, y

        result = self._convert_to_tensors(x, y)
        if isinstance(result, tuple):
            x_tensor, y_tensor = result
            x_np, y_np = x_tensor.cpu().numpy(), y_tensor.cpu().numpy()
            if (y_np.ndim == 2 and y_np.shape[1] == 1) or y_np.ndim == 1:
                y_np = y_np.ravel()  # Ensure y is 1-dimensional
                return check_X_y(x_np, y_np, multi_output=False, y_numeric=True)
            return check_X_y(x_np, y_np, multi_output=True, y_numeric=True)

        x_tensor = result
        return x_tensor.cpu().numpy(), None

    def _ensure_numpy_2d(self, arr: np.ndarray) -> np.ndarray:
        """Ensure that arr is a 2D numpy array."""
        if arr.ndim == 1:
            arr = arr.reshape(-1, 1)
        return arr

    def _random_split(
        self,
        dataset: Dataset,
        batch_size: int | None = None,
        train_size: float = 0.8,
        test_size: float = 0.2,
    ) -> tuple[DataLoader, DataLoader]:
        """
        Split Dataset into train/test DataLoaders.

        Parameters
        ----------
        dataset: Dataset
            The data to split.
        batch_size: int | None
            The DataLoader batch_size. If None, sets batch_size to lenth of training
            data. Defaults to None.
        """
        if train_size < 0.0 or train_size > 1.0 or test_size < 0.0 or test_size > 1.0:
            raise ValueError(
                f"Train size ({train_size}) and test size ({test_size}) must be "
                "specified as a proportion between 0 and 1"
            )
        if test_size + train_size != 1.0:
            raise ValueError(
                f"Train size ({train_size}) and test size ({test_size}) must sum to 1"
            )
        train, test = tuple(random_split(dataset, [train_size, test_size]))
        if batch_size is None:
            batch_size = len(train)
        train_loader = self._convert_to_dataloader(train, batch_size=batch_size)
        test_loader = self._convert_to_dataloader(test, batch_size=batch_size)
        return train_loader, test_loader

    @staticmethod
    def _normalize(x: TensorLike) -> tuple[TensorLike, TensorLike, TensorLike]:
        x_mean = x.mean(0, keepdim=True)
        x_std = x.std(0, keepdim=True)
        return (x - x_mean) / x_std, x_mean, x_std

    @staticmethod
    def _denormalize(
        x: TensorLike, x_mean: TensorLike, x_std: TensorLike
    ) -> TensorLike:
        return (x * x_std) + x_mean


def set_random_seed(seed: int = 42, deterministic: bool = True):
    """
    Set random seed for Python, NumPy and PyTorch.

    Parameters
    ----------
    seed: int
        The random seed to use.
    deterministic: bool
        Use "deterministic" algorithms in PyTorch.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    if deterministic:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)


class ValidationMixin:
    """
    Mixin class for validation methods.

    This class provides static methods for checking the types and shapes of
    input and output data, as well as validating specific tensor shapes.
    """

    @staticmethod
    def _check(x: TensorLike, y: TensorLike | None):
        """
        Check the types and shape are correct for the input data.

        Checks are equivalent to sklearn's check_array.
        """
        if not isinstance(x, TensorLike):
            raise ValueError(f"Expected x to be TensorLike, got {type(x)}")

        if y is not None and not isinstance(y, TensorLike):
            raise ValueError(f"Expected y to be TensorLike, got {type(y)}")

        # Check x
        if not torch.isfinite(x).all():
            msg = "Input tensor x contains non-finite values"
            raise ValueError(msg)
        if x.dtype not in TorchScalarDType:
            msg = (
                f"Input tensor x has unsupported dtype {x.dtype}. "
                "Expected float32, float64, int32, or int64."
            )
            raise ValueError(msg)

        # Check y if not None
        if y is not None:
            if not torch.isfinite(y).all():
                msg = "Input tensor y contains non-finite values"
                raise ValueError(msg)
            if y.dtype not in TorchScalarDType:
                msg = (
                    f"Input tensor y has unsupported dtype {y.dtype}. "
                    "Expected float32, float64, int32, or int64."
                )
                raise ValueError(msg)

        return x, y

    @staticmethod
    def _check_output(output: OutputLike):
        """Check the types and shape are correct for the output data."""
        if not isinstance(output, OutputLike):
            raise ValueError(f"Expected OutputLike, got {type(output)}")

        if isinstance(output, TensorLike) and output.ndim != 2:
            raise ValueError(f"Expected output to be 2D tensor, got {output.ndim}D")

    @staticmethod
    def check_vector(x: TensorLike) -> TensorLike:
        """
        Validate that the input is a 1D TensorLike.

        Parameters
        ----------
        x: TensorLike
            Input tensor to validate.

        Returns
        -------
        TensorLike
            Validated 1D tensor.

        Raises
        ------
        ValueError
            If x is not a TensorLike or is not 1-dimensional.
        """
        if not isinstance(x, TensorLike):
            raise ValueError(f"Expected TensorLike, got {type(x)}")
        if x.ndim != 1:
            raise ValueError(f"Expected 1D tensor, got {x.ndim}D")
        return x

    @staticmethod
    def check_matrix(x: TensorLike) -> TensorLike:
        """
        Validate that the input is a 2D TensorLike.

        Parameters
        ----------
        x: TensorLike
            Input tensor to validate.

        Returns
        -------
        TensorLike
            Validated 2D tensor.

        Raises
        ------
        ValueError
            If x is not a TensorLike or is not 2-dimensional.
        """
        if not isinstance(x, TensorLike):
            raise ValueError(f"Expected TensorLike, got {type(x)}")
        if x.ndim != 2:
            raise ValueError(f"Expected 2D tensor, got {x.ndim}D")
        return x

    @staticmethod
    def check_pair(x: TensorLike, y: TensorLike) -> tuple[TensorLike, TensorLike]:
        """
        Validate that two tensors have the same number of rows.

        Parameters
        ----------
        x: TensorLike
            First tensor.
        y: TensorLike
            Second tensor.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            The validated pair of tensors.

        Raises
        ------
        ValueError
            If x and y do not have the same number of rows.
        """
        if x.shape[0] != y.shape[0]:
            msg = "x and y must have the same number of rows"
            raise ValueError(msg)
        return x, y

    @staticmethod
    def check_covariance(y: TensorLike, Sigma: TensorLike) -> TensorLike:
        """
        Validate and return the covariance matrix.

        Parameters
        ----------
        y: TensorLike
            Output tensor.
        Sigma: TensorLike
            Covariance matrix, which may be full, diagonal, or a scalar per sample.

        Returns
        -------
        TensorLike
            Validated covariance matrix.

        Raises
        ------
        ValueError
            If Sigma does not have a valid shape relative to y.
        """
        if (
            Sigma.shape == (y.shape[0], y.shape[1], y.shape[1])
            or Sigma.shape == (y.shape[0], y.shape[1])
            or Sigma.shape == (y.shape[0],)
        ):
            return Sigma
        msg = "Invalid covariance matrix shape"
        raise ValueError(msg)

    @staticmethod
    def trace(Sigma: TensorLike, d: int) -> TensorLike:
        """
        Compute the trace of the covariance matrix (A-optimal design criterion).

        Parameters
        ----------
        Sigma: TensorLike
            Covariance matrix (full, diagonal, or scalar).
        d: int
            Dimension of the output.

        Returns
        -------
        TensorLike
            The computed trace value.

        Raises
        ------
        ValueError
            If Sigma does not have a valid shape.
        """
        if Sigma.dim() == 3 and Sigma.shape[1:] == (d, d):
            return torch.diagonal(Sigma, dim1=1, dim2=2).sum(dim=1).mean()
        if Sigma.dim() == 2 and Sigma.shape[1] == d:
            return Sigma.sum(dim=1).mean()
        if Sigma.dim() == 1:
            return d * Sigma.mean()
        raise ValueError(f"Invalid covariance matrix shape: {Sigma.shape}")

    @staticmethod
    def logdet(Sigma: TensorLike, dim: int) -> TensorLike:
        """
        Return the log-determinant of the covariance matrix.

        Compute the log-determinant of the covariance matrix (D-optimal design
        criterion).

        Parameters
        ----------
        Sigma: TensorLike
            Covariance matrix (full, diagonal, or scalar).
        dim: int
            Dimension of the output.

        Returns
        -------
        TensorLike
            The computed log-determinant value.

        Raises
        ------
        ValueError
            If Sigma does not have a valid shape.
        """
        if len(Sigma.shape) == 3 and Sigma.shape[1:] == (dim, dim):
            return torch.logdet(Sigma).mean()
        if len(Sigma.shape) == 2 and Sigma.shape[1] == dim:
            return torch.sum(torch.log(Sigma), dim=1).mean()
        if len(Sigma.shape) == 1:
            return dim * torch.log(Sigma).mean()
        raise ValueError(f"Invalid covariance matrix shape: {Sigma.shape}")

    @staticmethod
    def max_eigval(Sigma: TensorLike) -> TensorLike:
        """
        Return the maximum eigenvalue of the covariance matrix.

        Compute the maximum eigenvalue of the covariance matrix (E-optimal design
        criterion).

        Parameters
        ----------
        Sigma: TensorLike
            Covariance matrix (full, diagonal, or scalar).

        Returns
        -------
        TensorLike
            The average maximum eigenvalue.

        Raises
        ------
        ValueError
            If Sigma does not have a valid shape.
        """
        if Sigma.dim() == 3 and Sigma.shape[1:] == (Sigma.shape[1], Sigma.shape[1]):
            eigvals = torch.linalg.eigvalsh(Sigma)
            return eigvals[:, -1].mean()  # Eigenvalues are sorted in ascending order
        if Sigma.dim() == 2:
            return Sigma.max(dim=1).values.mean()
        if Sigma.dim() == 1:
            return Sigma.mean()
        raise ValueError(f"Invalid covariance matrix shape: {Sigma.shape}")

    ### Validation methods from old utils.py ###
    ### Leave here in case want to restore/refactor later ###

    # def _ensure_1d_if_column_vec(arr):
    #     """Ensure that arr is 1D if shape is (n, 1)."""
    #     if arr.ndim == 2 and arr.shape[1] == 1:
    #         arr = arr.ravel()
    #     if arr.ndim > 2 or arr.ndim < 1:
    #         raise ValueError(
    #             f"arr should be 1D or 2D. Found {arr.ndim}D array with shape {arr.shape}"  # noqa: E501
    #         )
    #     return arr

    # def _check_cv(cv):
    #     """Ensure that cross-validation method is valid"""
    #     if cv is None:
    #         msg = "cross_validator cannot be None"
    #         raise ValueError(msg)
    #     if not isinstance(cv, KFold):
    #         msg = (
    #             "cross_validator should be an instance of KFold cross-validation. We do not "  # noqa: E501
    #             "currently support other cross-validation methods."
    #         )
    #         raise ValueError(
    #             msg
    #         )
    #     return cv



import pandas as pd

from autoemulate.core.types import ModelParams
from autoemulate.emulators.transformed.base import TransformedEmulator


class Result:
    """Represents a single result of an emulator evaluation."""

    def __init__(  # noqa: PLR0913
        self,
        id: int,
        model_name: str,
        model: TransformedEmulator,
        params: ModelParams,
        r2_test: float,
        rmse_test: float,
        r2_train: float,
        rmse_train: float,
        r2_test_std: float,
        rmse_test_std: float,
        r2_train_std: float,
        rmse_train_std: float,
    ):
        """Initialize a Result object.

        Parameters
        ----------
        id: int
            Unique identifier for the result.
        model_name: str
            Name of the model used in the evaluation.
        model: TransformedEmulator
            The emulator model used for predictions.
        params: ModelParams
            Parameters used for the model.
        r2_test: float
            R2 score on the test set.
        rmse_test: float
            Root Mean Squared Error on the test set.
        r2_train: float
            R2 score on the training set.
        rmse_train: float
            Root Mean Squared Error on the training set.
        r2_test_std: float
            Standard deviation of the R2 score on the test set.
        rmse_test_std: float
            Standard deviation of the RMSE on the test set.
        r2_train_std: float
            Standard deviation of the R2 score on the training set.
        rmse_train_std: float
            Standard deviation of the RMSE on the training set.

        """
        self.id = id
        self.model_name = model_name
        self.model = model
        self.x_transforms = model.x_transforms
        self.y_transforms = model.y_transforms
        self.params = params
        self.r2_test = r2_test
        self.rmse_test = rmse_test
        self.r2_test_std = r2_test_std
        self.rmse_test_std = rmse_test_std
        self.r2_train = r2_train
        self.rmse_train = rmse_train
        self.r2_train_std = r2_train_std
        self.rmse_train_std = rmse_train_std

    def metadata_df(self) -> pd.DataFrame:
        """
        Return the Result object as a dataframe, without the model.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
            ['id', 'model_name', 'x_transforms', 'y_transforms', 'params',
            'r2_test', 'rmse_test', 'r2_test_std', 'rmse_test_std',
            'r2_train', 'rmse_train', 'r2_train_std', 'rmse_train_std'].
        """

        # Serialize the params dictionary to a string representation
        def serialize_params(params):
            out = {}
            for k, v in params.items():
                if callable(v) or isinstance(v, type):
                    out[k] = v.__name__
                else:
                    out[k] = v
            return out

        return pd.DataFrame(
            {
                "id": [self.id],
                "model_name": [self.model_name],
                "x_transforms": [self.x_transforms],
                "y_transforms": [self.y_transforms],
                "params": str(serialize_params(self.params)),
                "r2_test": [self.r2_test],
                "rmse_test": [self.rmse_test],
                "r2_test_std": [self.r2_test_std],
                "rmse_test_std": [self.rmse_test_std],
                "r2_train": [self.r2_train],
                "rmse_train": [self.rmse_train],
                "r2_train_std": [self.r2_train_std],
                "rmse_train_std": [self.rmse_train_std],
            }
        )


class Results:
    """Container for multiple Result objects."""

    def __init__(
        self,
        results: list[Result] | None = None,
    ):
        """Initialize a Results object.

        Parameters
        ----------
        results: list[Result] | None
            A list of Result objects. If None, an empty list is created.
        """
        if results is None:
            results = []
        self.results = results
        self._id_to_result = {result.id: result for result in self.results}

    def _update_index(self):
        self._id_to_result = {result.id: result for result in self.results}

    def add_result(self, result: Result):
        """Add a Result object to the collection."""
        self.results.append(result)
        self._update_index()

    def summarize(self) -> pd.DataFrame:
        """
        Summarize the results in a DataFrame.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                ['model', 'x_transforms', 'y_transforms', 'params', 'r2_score',
                'rmse_score'].
        TODO: include test data
        """
        data = {
            "model_name": [result.model_name for result in self.results],
            "x_transforms": [result.x_transforms for result in self.results],
            "y_transforms": [result.y_transforms for result in self.results],
            "params": [result.params for result in self.results],
            "rmse_test": [result.rmse_test for result in self.results],
            "r2_test": [result.r2_test for result in self.results],
            "r2_test_std": [result.r2_test_std for result in self.results],
            "r2_train": [result.r2_train for result in self.results],
            "r2_train_std": [result.r2_train_std for result in self.results],
        }
        df = pd.DataFrame(data)
        return df.sort_values(by="r2_test", ascending=False)

    summarise = summarize

    def best_result(self) -> Result:
        """
        Get the model with the best result based on the highest R2 score.

        Returns
        -------
        Result
            The result with the highest R2 score.
        """
        if not self.results:
            msg = "No results available. Please run AutoEmulate.compare() first."
            raise ValueError(msg)
        return max(self.results, key=lambda r: r.r2_test)

    def get_result(self, result_id: int) -> Result:
        """
        Get a result by its ID.

        Parameters
        ----------
        result_id: int
            The ID of the model to retrieve.

        Returns
        -------
        Result
            The result with the specified ID.
        """
        try:
            return self._id_to_result[result_id]
        except KeyError as err:
            raise ValueError(f"No result found with ID: {result_id!s}") from err



from torch import nn

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.data.utils import set_random_seed
from autoemulate.transforms.standardize import StandardizeTransform

from ..base import DropoutTorchBackend


class MLP(DropoutTorchBackend):
    """
    Multi-Layer Perceptron (MLP) emulator.

    MLP provides a simple deterministic emulator with optional model stochasticity
    provided by different weight initialization and dropout.
    """

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = True,
        standardize_y: bool = True,
        activation_cls: type[nn.Module] = nn.ReLU,
        loss_fn_cls: type[nn.Module] = nn.MSELoss,
        epochs: int = 100,
        layer_dims: list[int] | None = None,
        weight_init: str = "default",
        scale: float = 1.0,
        bias_init: str = "default",
        dropout_prob: float | None = None,
        lr: float = 1e-2,
        random_seed: int | None = None,
        device: DeviceLike | None = None,
        **kwargs,
    ):
        """
        Multi-Layer Perceptron (MLP) emulator.

        MLP provides a simple deterministic emulator with optional model stochasticity
        provided by different weight initialization and dropout.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike
            Target values.
        activation_cls: type[nn.Module]
            Activation function to use in the hidden layers. Defaults to `nn.ReLU`.
        layer_dims: list[int] | None
            Dimensions of the hidden layers. If None, defaults to [32, 16].
            Defaults to None.
        weight_init: str
            Weight initialization method. Options are "default", "normal", "uniform",
            "zeros", "ones", "xavier_uniform", "xavier_normal", "kaiming_uniform",
            "kaiming_normal". Defaults to "default".
        scale: float
            Scale parameter for weight initialization methods. Used as:
            - gain for Xavier methods
            - std for normal distribution
            - bound for uniform distribution (range: [-scale, scale])
            - ignored for Kaiming methods (uses optimal scaling)
            Defaults to 1.0.
        bias_init: str
            Bias initialization method. Options: "zeros", "default":
                - "zeros" initializes biases to zero
                - "default" uses PyTorch's default uniform initialization
        dropout_prob: float | None
            Dropout probability for regularization. If None, no dropout is applied.
            Defaults to None.
        lr: float
            Learning rate for the optimizer. Defaults to 1e-2.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set. Defaults to None.
        device: DeviceLike | None
            Device to run the model on (e.g., "cpu", "cuda", "mps"). Defaults to None.
        **kwargs: dict
            Additional keyword arguments.

        Raises
        ------
        ValueError
            If the input dimensions of `x` and `y` are not matrices.
        """
        TorchDeviceMixin.__init__(self, device=device)
        nn.Module.__init__(self)

        if random_seed is not None:
            set_random_seed(seed=random_seed)

        # Ensure x and y are tensors with correct dimensions
        x, y = self._convert_to_tensors(x, y)

        # Construct the MLP layers
        layer_dims = [x.shape[1], *layer_dims] if layer_dims else [x.shape[1], 32, 16]
        layers = []
        for idx, dim in enumerate(layer_dims[1:]):
            layers.append(nn.Linear(layer_dims[idx], dim, device=self.device))
            layers.append(activation_cls())
            if dropout_prob is not None:
                layers.append(nn.Dropout(p=dropout_prob))

        # Add final layer without activation
        num_tasks = y.shape[1]
        layers.append(nn.Linear(layer_dims[-1], num_tasks, device=self.device))
        self.nn = nn.Sequential(*layers)

        # Finalize initialization
        self._initialize_weights(weight_init, scale, bias_init)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.epochs = epochs
        self.loss_fn = loss_fn_cls()
        self.lr = lr
        self.optimizer = self.optimizer_cls(self.nn.parameters(), lr=self.lr)  # type: ignore[call-arg] since all optimizers include lr
        self.scheduler_setup(kwargs)
        self.to(device)

    def forward(self, x):
        """Forward pass for the MLP."""
        return self.nn(x)

    @staticmethod
    def is_multioutput() -> bool:
        """MLP supports multi-output."""
        return True

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        scheduler_params = MLP.scheduler_params()
        return {
            "epochs": [100, 200],
            "layer_dims": [[8, 4], [16, 8], [32, 16], [64, 32, 16]],
            "lr": [5e-1, 2e-1, 1e-1, 1e-2, 1e-3],
            "batch_size": [16, 32],
            "weight_init": ["default", "normal"],
            "scale": [0.1, 1.0],
            "bias_init": ["default", "zeros"],
            "dropout_prob": [0.3, None],
            "scheduler_cls": scheduler_params["scheduler_cls"],
            "scheduler_kwargs": scheduler_params["scheduler_kwargs"],
        }



from collections.abc import Callable

import numpy as np
import scipy as sp
import torch

from autoemulate.core.types import TensorLike
from autoemulate.simulations.base import Simulator


class FlowProblem(Simulator):
    """
    Flow Problem Simulator.

    The system simulated is a tube with an input flow rate at any given time.
    The tube is divided to `ncomp` compartments which allows for the study of
    the pressure and flow rate at various points in the tube.
    """

    def __init__(
        self,
        parameters_range: dict[str, tuple[float, float]] | None = None,
        output_names: list[str] | None = None,
        log_level: str = "progress_bar",
        ncycles: int = 10,
        ncomp: int = 10,
    ):
        """
        Initialize the FlowProblem simulator.

        Parameters
        ----------
        parameters_range: dict[str, tuple[float, float]]
            Dictionary mapping input parameter names to their (min, max) ranges.
        output_names: list[str]
            List of output parameters' names.
        log_level: str
            Logging level for the simulator. Can be one of:
            - "progress_bar": shows a progress bar during batch simulations
            - "debug": shows debug messages
            - "info": shows informational messages
            - "warning": shows warning messages
            - "error": shows error messages
            - "critical": shows critical messages
        ncycles: int
            Number of cycles to simulate.
        ncomp: int
            Number of compartments in the tube.
        """
        if parameters_range is None:
            parameters_range = {
                # Cardiac cycle period (s)
                "T": (0.5, 2.0),
                # Pulse duration (s)
                "td": (0.1, 0.5),
                # Amplitude (e.g., pressure or flow rate)
                "amp": (100.0, 1000.0),
                # Time step (s)
                "dt": (0.0001, 0.01),
                # Compliance (unit varies based on context)
                "C": (20.0, 60.0),
                # Resistance (unit varies based on context)
                "R": (0.01, 0.1),
                # Inductance (unit varies based on context)
                "L": (0.001, 0.005),
                # Outflow resistance (unit varies based on context)
                "R_o": (0.01, 0.05),
                # Initial pressure (unit varies based on context)
                "p_o": (5.0, 15.0),
            }
        if output_names is None:
            output_names = ["pressure"]
        super().__init__(parameters_range, output_names, log_level)
        self.ncycles = ncycles
        self.ncomp = ncomp

    def _forward(self, x: TensorLike) -> TensorLike:
        """
        Calculate the pressure and flow rate in the tube compartments.

        Parameters
        ----------
            x: TensorLike
                Tensor of input parameter values to simulate. Will be converted to a
                dictionary internally.
                    - `T` (float): cycle length
                    - `td` (float): pulse duration, make sure to make this less than T
                    - `amp` (float): inflow amplitude
                    - `dt` (float): temporal discretisation resolution
                    - `C` (float): tube average compliance
                    - `R` (float): tube average impedance
                    - `L` (float): hydraulic impedance, inertia
                    - `R_o` (float) : outflow resistance
                    - `p_o` (float) : outflow pressure

        Returns
        -------
            TensorLike
                Presssure in the tube compartments at the end of the simulation.
        """
        # Convert tensor input to dictionary
        x_values = x.squeeze().tolist() if hasattr(x, "squeeze") else x

        # Create parameter dictionary
        params = {}
        for i, param_name in enumerate(self.param_names):
            value = x_values[i]
            params[param_name] = value

        def dfdt_fd(params_dict, t: float, y: np.ndarray, Q_in):
            Cn = params_dict["C"] / self.ncomp
            Rn = params_dict["R"] / self.ncomp
            Ln = params_dict["L"] / self.ncomp

            out = np.zeros((self.ncomp, 2))
            y_temp = y.reshape((-1, 2))

            for i in range(self.ncomp):
                if i > 0:
                    out[i, 0] = (y_temp[i - 1, 1] - y_temp[i, 1]) / Cn
                else:
                    out[i, 0] = (Q_in(t % params_dict["T"]) - y_temp[i, 1]) / Cn
                if i < self.ncomp - 1:
                    out[i, 1] = (
                        -y_temp[i + 1, 0] + y_temp[i, 0] - Rn * y_temp[i, 1]
                    ) / Ln
                else:
                    out[i, 1] = (
                        -params_dict["p_o"]
                        + y_temp[i, 0]
                        - (Rn + params_dict["R_o"]) * y_temp[i, 1]
                    ) / Ln
            return out.reshape((-1,))

        def generate_pulse_function(params_dict: dict) -> Callable:
            def Q_mi_lambda(t):
                return (
                    np.sin(np.pi / params_dict["td"] * t) ** 2.0
                    * np.heaviside(params_dict["td"] - t, 0.0)
                    * params_dict["amp"]
                )

            return Q_mi_lambda

        def dfdt_fd_spec(t, y):
            return dfdt_fd(
                params_dict=params, t=t, y=y, Q_in=generate_pulse_function(params)
            )

        res = sp.integrate.solve_ivp(
            dfdt_fd_spec,
            [0.0, params["T"] * self.ncycles],
            y0=np.zeros(self.ncomp * 2),
            method="BDF",
            max_step=params["dt"],
        )
        res.y = res.y[:, res.t >= params["T"] * (self.ncycles - 1)]
        # Get the peak pressure across the
        # first ncomp compartments across all time steps
        peak_pressure = res.y[: self.ncomp, :].max()
        return torch.tensor([[peak_pressure]], dtype=torch.float32)



from typing import Any, TypeAlias

import numpy as np
import torch
import torch.utils
import torch.utils.data
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from torch.utils.data import DataLoader

NumpyLike: TypeAlias = np.ndarray
TensorLike: TypeAlias = torch.Tensor
DistributionLike: TypeAlias = torch.distributions.Distribution
GaussianLike: TypeAlias = MultivariateNormal
GaussianProcessLike: TypeAlias = MultitaskMultivariateNormal
InputLike: TypeAlias = NumpyLike | TensorLike | DataLoader | torch.utils.data.Dataset
OutputLike: TypeAlias = DistributionLike | TensorLike
ParamLike: TypeAlias = Any
TuneParams: TypeAlias = dict[str, list[ParamLike]]
ModelParams: TypeAlias = dict[str, ParamLike]
DeviceLike: TypeAlias = str | torch.device

# Torch dtype's
TorchScalarDType = (torch.float32, torch.float64, torch.int32, torch.int64)



import ast
from pathlib import Path

import joblib
import pandas as pd

from autoemulate.core.results import Result  # , Results
from autoemulate.emulators.base import Emulator


class ModelSerialiser:
    """ModelSerialiser handles saving and loading of models and results."""

    def __init__(self, logger):
        self.logger = logger

    def _save_model(
        self,
        model: Emulator,
        model_name: str | None = None,
        path: str | Path | None = None,
    ):
        """Save a model to disk.

        Parameters
        ----------
        model : Emulator model
            Model to save.
        model_name : str or None, optional
            Name of the model to save.
            If None, the model's name will be used.
        path : str, Path or None, optional
            Path to save the model.
            If None, the model will be saved with the model name.
        """
        model_filename = self._get_model_filename(model, model_name)
        full_path = self._prepare_path(path, model_filename)
        try:
            if full_path.suffix != ".joblib":
                joblib.dump(model, full_path.with_suffix(".joblib"))
            else:
                joblib.dump(model, full_path)
            self.logger.info("%s saved to %s", model_filename, full_path)
            return full_path
        except Exception as e:
            self.logger.error(
                "Failed to save %s to %s: %s", model_filename, full_path, e
            )
            raise

    def _save_result(
        self,
        result: Result,
        result_name: str | None = None,
        path: str | Path | None = None,
    ) -> Path:
        """Save a result and model to disk.

        Parameters
        ----------
        result : Result
            Result object containing the model and its metadata.
        result_name : str or None, optional
            Name of the result to save. If None, the result name and id will be used.
        path : str, Path or None, optional
            Path to save the model and metadata.
            If None, the result will be saved with the result_name.

        Returns
        -------
        Path
            Full path of saved files without either the .joblib or _metadata.csv suffix.
        """
        if result_name is None:
            result_name = f"{result.model_name}_{result.id}"
        full_path = self._prepare_path(path, result_name)

        self._save_model(result.model, result_name, full_path)

        # Save metadata to CSV
        metadata_path = Path(f"{full_path}_metadata.csv")
        metadata_df = result.metadata_df()
        try:
            metadata_df.to_csv(metadata_path, index=False)
            self.logger.info("Metadata saved to %s", metadata_path)
        except Exception as e:
            self.logger.error("Failed to save metadata to %s: %s", metadata_path, e)
            raise

        return full_path

    def _load_model(self, path: str | Path):
        """Load a model from disk.

        Parameters
        ----------
        path : str or Path
            Path to load the model.
        """
        path = Path(path)
        try:
            model = joblib.load(path)
            self.logger.info("Model loaded from %s", path)
            return model
        except Exception as e:
            self.logger.error("Failed to load model from %s: %s", path, e)
            raise

    def _load_result(self, path: str | Path) -> Result | Emulator:
        """
        Load a result or model from disk.

        Load a model and (if it exists) its metadata from disk,
        returning either a Result or Emulator object.

        Parameters
        ----------
        path : str or Path
            Path to the model file.

        Returns
        -------
        Result or Emulator
            The loaded model or result object.
        """
        if ".joblib" not in str(path):
            model_path = Path(f"{path}.joblib")
        else:
            model_path = Path(path)
        model = self._load_model(model_path)
        metadata_path = Path(f"{path}_metadata.csv")
        try:
            metadata_df = pd.read_csv(metadata_path, nrows=1)
            self.logger.info("Metadata loaded from %s", metadata_path)
        except Exception as e:
            msg = "Failed to load metadata from %s: %s"
            self.logger.error(msg, metadata_path, e)
            return model
        row = metadata_df.iloc[0]
        params = row["params"]
        params = ast.literal_eval(params)

        return Result(
            id=row["id"],
            model_name=row["model_name"],
            model=model,
            params=params,
            r2_test=row["r2_test"],
            rmse_test=row["rmse_test"],
            r2_test_std=row["r2_test_std"],
            rmse_test_std=row["rmse_test_std"],
            r2_train=row["r2_train"],
            rmse_train=row["rmse_train"],
            r2_train_std=row["r2_train_std"],
            rmse_train_std=row["rmse_train_std"],
        )

    def _prepare_path(self, path: str | Path | None, model_name: str):
        """Prepare path for saving model."""
        if path is None:
            full_path = Path(model_name)
        else:
            full_path = Path(path)
            if full_path.is_dir():
                full_path = full_path / model_name

        full_path.parent.mkdir(parents=True, exist_ok=True)
        return full_path

    @staticmethod
    def _get_model_filename(model, provided_name: str | None = None) -> str:
        """Get the name of the model to save."""
        model_name = provided_name if provided_name is not None else model.model_name()
        return f"{model_name}.joblib"



import warnings
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tqdm

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.logging_config import get_configured_logger
from autoemulate.core.model_selection import bootstrap, evaluate, r2_metric
from autoemulate.core.plotting import (
    calculate_subplot_layout,
    display_figure,
    plot_xy,
)
from autoemulate.core.results import Result, Results
from autoemulate.core.save import ModelSerialiser
from autoemulate.core.tuner import Tuner
from autoemulate.core.types import DeviceLike, DistributionLike, InputLike, ModelParams
from autoemulate.data.utils import ConversionMixin, set_random_seed
from autoemulate.emulators import (
    ALL_EMULATORS,
    get_emulator_class,
)
from autoemulate.emulators.base import Emulator
from autoemulate.emulators.transformed.base import TransformedEmulator
from autoemulate.transforms.base import AutoEmulateTransform
from autoemulate.transforms.standardize import StandardizeTransform


class AutoEmulate(ConversionMixin, TorchDeviceMixin, Results):
    """
    Automated emulator fitting.

    The AutoEmulate class is the main class of the AutoEmulate package.
    It is used to set up and compare different emulator models on a given dataset.
    It can also be used to summarise and visualise results, and to save and load models.

    """

    def __init__(  # noqa: PLR0913
        self,
        x: InputLike,
        y: InputLike,
        models: list[type[Emulator] | str] | None = None,
        x_transforms_list: list[list[AutoEmulateTransform | dict]] | None = None,
        y_transforms_list: list[list[AutoEmulateTransform | dict]] | None = None,
        model_tuning: bool = True,
        model_kwargs: None | ModelParams = None,
        n_iter: int = 10,
        n_splits: int = 5,
        shuffle: bool = True,
        n_bootstraps: int = 100,
        max_retries: int = 3,
        device: DeviceLike | None = None,
        random_seed: int | None = None,
        log_level: str = "progress_bar",
    ):
        """
        Initialize the AutoEmulate class.

        Parameters
        ----------
        x: InputLike
            Input features.
        y: InputLike or None
            Target values (not needed if x is a Dataset).
        models: list[type[Emulator]] | None
            List of emulator classes to compare. If None, all available emulators
            are used.
        x_transforms_list: list[list[AutoEmulateTransform]] | None
            An optional list of sequences of transforms to apply to the input data.
            Defaults to None, in which case the data is standardized.
        y_transforms_list: list[list[AutoEmulateTransform]] | None
            An optional list of sequences of transforms to apply to the output data.
            Defaults to None, in which case the data is standardized.
        model_tuning: bool
            Whether to tune the hyperparameters of the models using cross-validation.
            If False, the models use the model_kwargs provided. Defaults to True.
        model_kwargs: ModelParams | None
            Dictionary of model-specific parameters to use when fitting the models.
            If None, the default parameters for each model are used.
            This is only used if model_tuning is False. Defaults to None.
        n_iter: int
            Number of parameter settings to randomly sample and test during tuning.
        n_splits: int
            Number of cross validation folds to split data into. Defaults to 5.
        shuffle: bool
            Whether to shuffle data before splitting into cross validation folds.
            Defaults to True.
        n_bootstraps: int
            Number of times to resample the data when evaluating performance.
        device: DeviceLike | None
            Device to run the emulators. If None, uses the default device (usually CPU
            or GPU). Defauls to None.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set. Defaults to None.
        log_level: str
            Logging level. Can be "progress_bar", "debug", "info", "warning",
            "error", or "critical". Defaults to "progress_bar". If "progress_bar",
            it will show a progress bar during model comparison. It will set the
            logging level to "error" to avoid cluttering the output
            with debug/info logs.
        """
        Results.__init__(self)
        self.random_seed = random_seed
        TorchDeviceMixin.__init__(self, device=device)
        x, y = self._convert_to_tensors(x, y)
        x, y = self._move_tensors_to_device(x, y)

        # Transforms to search over
        self.x_transforms_list = [
            self.get_transforms(transforms)
            for transforms in (x_transforms_list or [[StandardizeTransform()]])
        ]
        self.y_transforms_list = [
            self.get_transforms(transforms)
            for transforms in (y_transforms_list or [[StandardizeTransform()]])
        ]

        # Set default models if None
        updated_models = self.get_models(models)

        # Filter models to only be those that can handle multioutput data
        if y.shape[1] > 1:
            updated_models = self.filter_models_if_multioutput(
                updated_models, models is not None
            )

        self.models = updated_models
        if random_seed is not None:
            set_random_seed(seed=random_seed)
        self.train_val, self.test = self._random_split(self._convert_to_dataset(x, y))

        # Run the compare method with the provided models
        if not self.models:
            msg = (
                "No models provided or available for comparison. "
                "Please provide a list of models to compare."
            )
            raise ValueError(msg)

        # Assign tuner parameters
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.n_iter = n_iter
        self.n_bootstraps = n_bootstraps
        self.max_retries = max_retries
        self.model_tuning = model_tuning
        self.model_kwargs = model_kwargs or {}

        # Set up logger and ModelSerialiser for saving models
        self.logger, self.progress_bar = get_configured_logger(log_level)
        self.model_serialiser = ModelSerialiser(self.logger)

        # Run compare
        self.compare()

    @staticmethod
    def all_emulators() -> list[type[Emulator]]:
        """Return a list of all available emulators."""
        return ALL_EMULATORS

    @staticmethod
    def list_emulators() -> pd.DataFrame:
        """Return a dataframe with model names of all available emulators.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns ['model_name', 'short_name'].
        """
        return pd.DataFrame(
            {
                "Emulator": [emulator.model_name() for emulator in ALL_EMULATORS],
                # TODO: short_name not currently used for anything, so commented out
                # "short_name": [emulator.short_name() for emulator in ALL_EMULATORS],
            }
        )

    def get_models(
        self, models: list[type[Emulator] | str] | None = None
    ) -> list[type[Emulator]]:
        """Return a list of the model classes for comparisons."""
        if models is None:
            return self.all_emulators()

        model_classes = []
        for model in models:
            if isinstance(model, str):
                model_classes.append(get_emulator_class(model))
            elif issubclass(model, Emulator):
                model_classes.append(model)
            else:
                raise ValueError(
                    f"Invalid model type: {type(model)}. "
                    "Expected a string or a subclass of Emulator"
                )
        return model_classes

    def get_transforms(
        self, transforms: list[AutoEmulateTransform | dict[str, object]]
    ) -> list[AutoEmulateTransform]:
        """Process and return a list of transforms."""
        processed_transforms = []
        for transform in transforms:
            if isinstance(transform, dict):
                deserialized_transform = AutoEmulateTransform.from_dict(transform)
                processed_transforms.append(deserialized_transform)
            elif isinstance(transform, AutoEmulateTransform):
                processed_transforms.append(transform)
            else:
                msg = f"Invalid transform type: {type(transform)}"
                raise ValueError(msg)
        return processed_transforms

    def filter_models_if_multioutput(
        self, models: list[type[Emulator]], warn: bool
    ) -> list[type[Emulator]]:
        """Filter models to only include those that support multi-output data."""
        updated_models = []
        for model in models:
            if not model.is_multioutput():
                if warn:
                    msg = (
                        f"Model ({model}) is not multioutput but the data is "
                        f"multioutput. Skipping model ({model})..."
                    )
                    warnings.warn(msg, stacklevel=2)
            else:
                updated_models.append(model)
        return updated_models

    def log_compare(  # noqa: PLR0913
        self,
        best_model_name,
        x_transforms,
        y_transforms,
        best_params_for_this_model,
        r2_score,
        rmse_score,
    ):
        """Log the comparison results."""
        msg = (
            "Comparison results:\n"
            f"Best Model: {best_model_name}, "
            f"x transforms: {x_transforms}, "
            f"y transforms: {y_transforms}",
            f"Best params: {best_params_for_this_model}, "
            f"R2 score: {r2_score:.3f}, "
            f"RMSE score: {rmse_score:.3f}",
        )
        self.logger.debug(msg)

    def compare(self):
        """
        Compare different models on the provided dataset.

        The method will:
        - Loop over all combinations of x and y transforms and models.
        - Set up the tuner with the training/validation data.
        - Tune hyperparameters for each model.
        - Fit the best model with the tuned hyperparameters.
        - Evaluate the performance of the best model on the test data.
        - Log the results.
        - Save the best model and its parameters.
        """
        tuner = Tuner(self.train_val, y=None, n_iter=self.n_iter, device=self.device)
        self.logger.info(
            "Comparing %s", [model_cls.__name__ for model_cls in self.models]
        )
        for x_transforms in self.x_transforms_list:
            for y_transforms in self.y_transforms_list:
                for id, model_cls in tqdm.tqdm(
                    enumerate(self.models),
                    disable=not self.progress_bar,
                    desc="Comparing models",
                    total=len(self.models),
                    unit="model",
                    unit_scale=True,
                ):
                    # Since refitting the model can fail after tuning, max_retries
                    # is used to retry the whole tuning and fitting process
                    for attempt in range(self.max_retries):
                        try:
                            self.logger.info(
                                "Running Model: %s: %d/%d (attempt %d/%d)",
                                model_cls.__name__,
                                id + 1,
                                len(self.models),
                                attempt + 1,
                                self.max_retries,
                            )
                            if self.model_tuning:
                                self.logger.debug(
                                    'Running tuner for model "%s"',
                                    model_cls.__name__,
                                )
                                scores, params_list = tuner.run(
                                    model_cls,
                                    x_transforms,
                                    y_transforms,
                                    n_splits=self.n_splits,
                                    shuffle=self.shuffle,
                                )
                                mean_scores = [
                                    np.mean(score).item() for score in scores
                                ]
                                best_score_idx = np.argmax(mean_scores)
                                best_params_for_this_model = params_list[best_score_idx]
                                self.logger.debug(
                                    'Tuner found best params for model "%s": '
                                    "%s with score: %.3f",
                                    model_cls.__name__,
                                    best_params_for_this_model,
                                    mean_scores[best_score_idx],
                                )
                            else:
                                self.logger.debug(
                                    'Skipping tuning for model "%s", using default'
                                    "parameters",
                                    model_cls.__name__,
                                )
                                best_params_for_this_model = self.model_kwargs

                            self.logger.debug(
                                'Running cross-validation for model "%s" '
                                'for "%s" iterations',
                                model_cls.__name__,
                                self.n_iter,
                            )
                            train_val_x, train_val_y = self._convert_to_tensors(
                                self.train_val
                            )
                            test_x, test_y = self._convert_to_tensors(self.test)
                            transformed_emulator = TransformedEmulator(
                                train_val_x,
                                train_val_y,
                                model=model_cls,
                                x_transforms=x_transforms,
                                y_transforms=y_transforms,
                                device=self.device,
                                **best_params_for_this_model,
                            )

                            # This can fail for some model params
                            transformed_emulator.fit(train_val_x, train_val_y)

                            (
                                (r2_train_val, r2_train_val_std),
                                (rmse_train_val, rmse_train_val_std),
                            ) = bootstrap(
                                transformed_emulator,
                                train_val_x,
                                train_val_y,
                                n_bootstraps=self.n_bootstraps,
                                device=self.device,
                            )
                            (r2_test, r2_test_std), (rmse_test, rmse_test_std) = (
                                bootstrap(
                                    transformed_emulator,
                                    test_x,
                                    test_y,
                                    n_bootstraps=self.n_bootstraps,
                                    device=self.device,
                                )
                            )

                            self.logger.debug(
                                'Cross-validation for model "%s"'
                                " completed with test mean (std) R2 score: %.3f (%.3f),"
                                " mean (std) RMSE score: %.3f (%.3f)",
                                model_cls.__name__,
                                r2_test,
                                r2_test_std,
                                rmse_test,
                                rmse_test_std,
                            )
                            self.logger.info(
                                "Finished running Model: %s\n", model_cls.__name__
                            )
                            result = Result(
                                id=id,
                                model_name=transformed_emulator.untransformed_model_name,
                                model=transformed_emulator,
                                params=best_params_for_this_model,
                                r2_test=r2_test,
                                rmse_test=rmse_test,
                                r2_test_std=r2_test_std,
                                rmse_test_std=rmse_test_std,
                                r2_train=r2_train_val,
                                rmse_train=rmse_train_val,
                                r2_train_std=r2_train_val_std,
                                rmse_train_std=rmse_train_val_std,
                            )
                            self.add_result(result)
                            # if successful, break out of the retry loop
                            break
                        except Exception as e:
                            self.logger.warning(
                                "Model %s failed on attempt %d/%d: %s",
                                model_cls.__name__,
                                attempt + 1,
                                self.max_retries,
                                str(e),
                            )
                            if attempt == self.max_retries - 1:
                                self.logger.error(
                                    "Model %s failed after %d attempts, skipping.",
                                    model_cls.__name__,
                                    self.max_retries,
                                )

        # Get the best result and log the comparison
        best_result = self.best_result()
        self.log_compare(
            best_model_name=best_result.model_name,
            x_transforms=best_result.x_transforms,
            y_transforms=best_result.y_transforms,
            best_params_for_this_model=best_result.params,
            r2_score=best_result.r2_test,
            rmse_score=best_result.rmse_test,
        )

    def plot(  # noqa: PLR0912, PLR0915
        self,
        model_obj: int | Emulator | Result,
        input_index: list[int] | int | None = None,
        output_index: list[int] | int | None = None,
        figsize=None,
        n_cols: int = 3,
    ):
        """
        Plot the evaluation of the model with the given result_id.

        Parameters
        ----------
        model_obj: int | Emulator | Result
            The model to plot. Can be an integer ID of a Result, an Emulator instance,
            or a Result instance.
        input_index: int
            The index of the input feature to plot against the output.
        output_index: int
            The index of the output feature to plot against the input.
        """
        result = None
        if isinstance(model_obj, int):
            if model_obj not in self._id_to_result:
                raise ValueError(f"No result found with ID: {model_obj}")
            result = self.get_result(model_obj)
            model = result.model
        elif isinstance(model_obj, Emulator):
            model = model_obj
        elif isinstance(model_obj, Result):
            model = model_obj.model

        test_x, test_y = self._convert_to_tensors(self.test)

        # Re-run prediction with just this model to get the predictions
        y_pred = model.predict(test_x)
        y_variance = None
        if isinstance(y_pred, DistributionLike):
            y_variance = y_pred.variance
            y_pred = y_pred.mean
        r2_score = evaluate(y_pred, test_y, r2_metric())

        # Convert to numpy arrays for plotting and ensure correct shapes
        test_x, test_y = self._convert_to_numpy(test_x, test_y)
        y_pred, _ = self._convert_to_numpy(y_pred, None)
        assert test_x is not None
        assert test_y is not None
        assert y_pred is not None
        test_x = self._ensure_numpy_2d(test_x)
        test_y = self._ensure_numpy_2d(test_y)
        y_pred = self._ensure_numpy_2d(y_pred)
        if y_variance is not None:
            y_variance, _ = self._convert_to_numpy(y_variance, None)
            y_variance = self._ensure_numpy_2d(y_variance)

        _, n_features = test_x.shape

        n_outputs = test_y.shape[1] if test_y.ndim > 1 else 1

        # Handle input and output indices
        if input_index is None:
            input_index = list(range(n_features))
        elif isinstance(input_index, int):
            input_index = [input_index]

        if output_index is None:
            output_index = list(range(n_outputs))
        elif isinstance(output_index, int):
            output_index = [output_index]

        # check that input_index and output_index are valid
        if any(idx >= n_features for idx in input_index):
            msg = f"input_index {input_index} is out of range. "
            msg += f"The index should be between 0 and {n_features - 1}."
            raise ValueError(msg)
        if any(idx >= n_outputs for idx in output_index):
            msg = f"output_index {output_index} is out of range. "
            msg += f"The index should be between 0 and {n_outputs - 1}."
            raise ValueError(msg)

        # Calculate number of subplots
        n_plots = len(input_index) * len(output_index)

        # Calculate number of rows
        n_rows, n_cols = calculate_subplot_layout(n_plots, n_cols)

        # Set up the figure
        if figsize is None:
            figsize = (5 * n_cols, 4 * n_rows)
        fig, axs = plt.subplots(n_rows, n_cols, figsize=figsize, squeeze=False)
        axs = axs.flatten()

        plot_index = 0
        for out_idx in output_index:
            for in_idx in input_index:
                if plot_index < len(axs):
                    plot_xy(
                        test_x[:, in_idx],
                        test_y[:, out_idx],
                        y_pred[:, out_idx],
                        y_variance[:, out_idx] if y_variance is not None else None,
                        ax=axs[plot_index],
                        title=f"$x_{in_idx}$ vs. $y_{out_idx}$",
                        input_index=in_idx,
                        output_index=out_idx,
                        r2_score=r2_score,
                    )
                    plot_index += 1

        # Hide any unused subplots
        for ax in axs[plot_index:]:
            ax.set_visible(False)
        plt.tight_layout()

        return display_figure(fig)

    def save(
        self,
        model_obj: int | Emulator | Result,
        path: str | Path | None = None,
        use_timestamp: bool = True,
    ) -> Path:
        """Save model to disk.

        Parameters
        ----------
        model_obj : int | Emulator | Result
            The model to save. Can be an integer ID of a Result, an Emulator instance,
            or a Result instance.
        path : str
            Path to save the model.
        use_timestamp : bool
            If True, appends a timestamp to the filename to ensure uniqueness.
        """
        result = None
        if isinstance(model_obj, int):
            if model_obj not in self._id_to_result:
                raise ValueError(f"No result found with ID: {model_obj}")
            result = self.get_result(model_obj)
            model = result.model
            model_name = result.model_name
        elif isinstance(model_obj, Emulator):
            model = model_obj
            if isinstance(model_obj, TransformedEmulator):
                model_name = model_obj.untransformed_model_name
            else:
                model_name = model.model_name()
        elif isinstance(model_obj, Result):
            model = model_obj.model
            model_name = model_obj.model_name
            result = model_obj

        # Create a unique filename based on the model name, id and date
        filename = f"{model_name}_{result.id}" if result is not None else model_name
        if use_timestamp:
            t = datetime.now().strftime("%Y%m%d_%H%M%S")
            filename += f"_{t}"

        if result is not None:
            return self.model_serialiser._save_result(result, filename, path)
        return self.model_serialiser._save_model(model, filename, path)

    def load(self, path: str | Path) -> Emulator | Result:
        """Load a stored model or result from disk.

        Parameters
        ----------
        path : str
            Path to model.

        Returns
        -------
        Emulator | Result
            The loaded model or result object.
        """
        return self.model_serialiser._load_result(path)
